from noofa.tests.base import NoofaTest
from noofa.core.sources import utils
from noofa.core.sources.query import (
    Q,
    Table,
    NoSuchFieldError,
    TableHasNoFields,
    SelectQuery,
    GtFilter,
    LtFilter,
    EqFilter,
    NeqFilter,
    GeFilter,
    LeFilter,
    EqFilter,
    ContainsFilter,
    StartsWithFilter,
    EndsWithFilter,
    InFilter,
    NotInFilter,
)


class TestQueryComponents(NoofaTest):
    """
    Тестирование компонентов запроса.
    """
    @classmethod
    def setUpClass(cls):
        cls.table = Table(
            'table1',
            ['col1', 'col2', 'col3'],
            enquote=True
        )
        cls.table2 = Table(
            'table2',
            ['col1', 'col2', 'col3'],
            enquote=True
        )
        cls.table_no_enquote = Table(
            'table1',
            ['col1', 'col2', 'col3'],
            enquote=False
        )

    def setUp(self):
        self.maxDiff = None
        self.query = self.table.select()

    def test_table_has_field(self):
        self.assertTrue(self.table.has_field('col1'))
        self.assertRaises(NoSuchFieldError, self.table.has_field, 'col4')

    def test_empty_table(self):
        args = ('table', [])
        self.assertRaises(TableHasNoFields, Table, *args)

    def test_table_verbose_names(self):
        vn = self.table.get_verbose_names()
        expected = ['"table1"."col1"', '"table1"."col2"', '"table1"."col3"']
        self.assertEqual(vn, expected)

    def test_table_verbose_names_no_enquote(self):
        vn = self.table_no_enquote.get_verbose_names()
        expected = ['table1.col1', 'table1.col2', 'table1.col3']
        self.assertEqual(vn, expected)

    def test_table_select(self):
        self.assertIsInstance(self.table.select(), SelectQuery)

    def test_select_query_str(self):
        select_query = self.query
        expected = 'SELECT "table1"."col1", "table1"."col2", "table1"."col3" FROM table1'
        self.assertEqual(str(select_query), expected)

    def test_select_query_join(self):
        join = self.table.columns.col1 == self.table2.columns.col2
        select_query = self.query
        select_query.join(join)

        expected = ''.join([
            'SELECT "table1"."col1", "table1"."col2", "table1"."col3", ',
            '"table2"."col1", "table2"."col2", "table2"."col3" FROM table1 ',
            'INNER JOIN "table2" ON "table1"."col1" = "table2"."col2"'
        ])
        self.assertEqual(str(select_query).rstrip(), expected)

    def test_select_query_filter(self):
        f = GtFilter('"table1"."col1"', 100, '%s')
        select_query = self.query
        select_query.where(f)
        q, params = select_query.str_and_params()

        expected_text = ''.join([
            'SELECT "table1"."col1", "table1"."col2", "table1"."col3" FROM table1 ',
            'WHERE "table1"."col1" > %s'
        ])
        self.assertEqual(q, expected_text)
        self.assertEqual(params, [100])

    def test_select_query_complex_filter(self):
        f1 = GtFilter('"table1"."col1"', 100, '%s')
        f2 = LtFilter('"table1"."col1"', 200, '%s')
        complex_fitler = Q(f1, f2)
        select_query = self.query
        select_query.where(complex_fitler)
        q, params = select_query.str_and_params()

        expected_text = ''.join([
            'SELECT "table1"."col1", "table1"."col2", "table1"."col3" FROM table1 ',
            'WHERE ("table1"."col1" > %s AND "table1"."col1" < %s)'
        ])
        self.assertEqual(q, expected_text)
        self.assertEqual(params, [100, 200])

    def test_select_query_limit(self):
        select_query = self.query
        expected_text = ''.join([
            'SELECT "table1"."col1", "table1"."col2", "table1"."col3" FROM table1 LIMIT 10',
        ])
        with self.subTest():
            for n in [10, '10', 10.5]:
                select_query.limit(n)
                self.assertEqual(str(select_query), expected_text)

    def test_select_query_order_by(self):
        select_query = self.query
        select_query.order_by((['"table1"."col1"'], 'asc'))

        expected_text = ''.join([
            'SELECT "table1"."col1", "table1"."col2", "table1"."col3" FROM table1 ',
            'ORDER BY "table1"."col1" ASC'
        ])
        self.assertEqual(str(select_query), expected_text)
        self.assertRaises(
            NoSuchFieldError,
            select_query.order_by,
            (['"table1"."col4"'], 'asc')
        )

    def test_empty_q_filter(self):
        q1 = Q()
        self.assertTrue(q1.is_empty)

        gt = GtFilter('"table1"."col1"', 100, '%s')
        q2 = Q(gt)
        expected_text = '"table1"."col1" > %s'
        self.assertEqual(str(q2), expected_text)

    def test_q_filter(self):
        def _test_or():
            Q() | 1
        self.assertRaises(TypeError, _test_or)

        gt = GtFilter('"table1"."col1"', 100, '%s')
        lt = LtFilter('"table1"."col1"', 200, '%s')
        q = Q(gt, lt)
        expected_text = '("table1"."col1" > %s AND "table1"."col1" < %s)'
        self.assertEqual(str(q), expected_text)

        q = Q(gt) | Q(lt)
        expected_text = '("table1"."col1" > %s OR "table1"."col1" < %s)'
        self.assertEqual(str(q), expected_text)


class TestQueryPreparation(NoofaTest):
    """
    Тестирование подготовки к построению запроса из json.
    """
    def setUp(self):
        self.base_query = {
            'base': 'table1',
            'tables': ['table1'],
        }
        self._subq = {
            'filters': [
                {
                    'is_complex': False, 'table': 'table1', 'field': 'f1', 'op': 'in', 'value': {
                        'is_subquery': True, 'base': 'table1', 'tables': ['table1'],
                        'values': [{'table': 'table1', 'field': 'f1'}],
                    }
                }
            ]
        }
        self._subq_another_table = {
            'filters': [
                {
                    'is_complex': False, 'table': 'table1', 'field': 'f1', 'op': 'in', 'value': {
                        'is_subquery': True, 'base': 'table2', 'tables': ['table2'],
                        'values': [{'table': 'table2', 'field': 'f2'}],
                    }
                }
            ]
        }
        self.with_subquery = {
            **self.base_query,
            **self._subq,
        }

    def test_no_subqueries(self):
        sq = utils._find_subqueries(self.base_query)
        sq_len = len(sq)
        self.assertEqual(sq_len, 0)

    def test_find_subqueries(self):
        sq = utils._find_subqueries(self.with_subquery)
        sq_len = len(sq)
        self.assertEqual(sq_len, 1)

    def test_find_tables(self):
        queries = utils._find_subqueries({**self.base_query, **self._subq_another_table})
        tables = utils._find_tables(queries)
        self.assertEqual(tables, ['table2'])

    def test_collect_tables(self):
        _ = {'base': 'table1', 'tables': ['table1', 'table3', 'table4']}
        tables = utils.collect_tables({**_, **self._subq_another_table})
        self.assertCountEqual(['table1', 'table2', 'table3', 'table4'], tables)


class TestQueryBuild(NoofaTest):
    """
    Тестирование построения запроса из json.
    """
    def setUp(self):
        self.maxDiff = None
        self.simple_json_query = {
            'base': 'table1',
            'tables': ['table1'],
        }
        self.values_simple = {
            'values': [{'table': 'table1', 'field': 'col1'}]
        }
        self.joins = {
            'joins': [
                {'l': 'table1', 'r': 'table2', 'j': 'inner',
                    'on': {'l': 'col1', 'r': 'col2'}},
                {'l': 'table1', 'r': 'table2', 'j': 'left',
                    'on': {'l': 'col2', 'r': 'col3'}},
                {'l': 'table1', 'r': 'table2', 'j': 'right',
                    'on': {'l': 'col3', 'r': 'col1'}},
            ]
        }
        self.limit = {'limit': 100}
        self.simple_filter = {
            'filters': [
                {'is_complex': False, 'table': 'table1', 'field': 'col1', 'op': '>', 'value': 100},
            ],
        }
        self.complex_filter = {
            'filters': [
                {'is_complex': True, 'op': 'and', 'filters': [
                    {'is_complex': False, 'table': 'table1', 'field': 'col1', 'op': '>', 'value': 100},
                    {'is_complex': False, 'table': 'table1', 'field': 'col1', 'op': '<', 'value': 200},
                ]},
            ],
        }
        self.orderby = {
            'order_by': [
                {'table': 'table1', 'fields': ['col1'], 'type': 'asc'},
            ],
        }
        self.mock_table1 = Table(
            'table1',
            ['col1', 'col2', 'col3'],
            enquote=True
        )
        self.mock_table2 = Table(
            'table2',
            ['col1', 'col2', 'col3'],
            enquote=True
        )

    def test_simple_select(self):
        qb = utils.Qbuilder(self.mock_tables, self.simple_json_query)
        query = qb.parse_query()
        expected_text = 'SELECT "table1"."col1", "table1"."col2", "table1"."col3" FROM table1'
        self.assertEqual(str(query), expected_text)

    def test_select_list(self):
        qb = utils.Qbuilder(self.mock_tables,
            {**self.simple_json_query, **self.values_simple})
        query = qb.parse_query()
        expected_text = 'SELECT "table1"."col1" FROM table1'
        self.assertEqual(str(query), expected_text)

    def test_join(self):
        qb = utils.Qbuilder(self.mock_tables,
            {**self.simple_json_query, **self.joins})
        query = qb.parse_query()
        expected_text = ''.join([
            'SELECT "table1"."col1", "table1"."col2", "table1"."col3",',
            ' "table2"."col1", "table2"."col2", "table2"."col3" ',
            'FROM table1 INNER JOIN "table2" ON "table1"."col1" = "table2"."col2" ',
            'LEFT OUTER JOIN "table2" ON "table1"."col2" = "table2"."col3" ',
            'RIGHT OUTER JOIN "table2" ON "table1"."col3" = "table2"."col1"'
        ])
        self.assertEqual(str(query), expected_text)

    def test_simple_filter(self):
        qb = utils.Qbuilder(
            self.mock_tables,
            {**self.simple_json_query, **self.simple_filter},
            param_placeholder='%s'
        )
        query = qb.parse_query()
        expected_text = ''.join([
            'SELECT "table1"."col1", "table1"."col2", "table1"."col3" FROM table1 ',
            'WHERE "table1"."col1" > %s'
        ])
        q, params = query.str_and_params()
        self.assertEqual(q, expected_text)
        self.assertEqual(params, [100, ])

    def test_complex_filter(self):
        qb = utils.Qbuilder(
            self.mock_tables,
            {**self.simple_json_query, **self.complex_filter},
            param_placeholder='%s'
        )
        query = qb.parse_query()
        expected_text = ''.join([
            'SELECT "table1"."col1", "table1"."col2", "table1"."col3" FROM table1 ',
            'WHERE ("table1"."col1" > %s AND "table1"."col1" < %s)'
        ])
        q, params = query.str_and_params()
        self.assertEqual(q, expected_text)
        self.assertEqual(params, [100, 200])

    def test_limit(self):
        qb = utils.Qbuilder(self.mock_tables,
            {**self.simple_json_query, **self.limit})
        query = qb.parse_query()
        expected_text = 'SELECT "table1"."col1", "table1"."col2", "table1"."col3" FROM table1 LIMIT 100'
        self.assertEqual(str(query), expected_text)

    def test_limit_mssql(self):
        qb = utils.Qbuilder(self.mock_tables,
            {**self.simple_json_query, **self.limit},
            mssql=True,
        )
        query = qb.parse_query()
        expected_text = 'SELECT TOP 100 "table1"."col1", "table1"."col2", "table1"."col3" FROM table1'
        self.assertEqual(str(query), expected_text)

    def test_orderby(self):
        qb = utils.Qbuilder(self.mock_tables,
            {**self.simple_json_query, **self.orderby})
        query = qb.parse_query()
        expected_text = ''.join([
            'SELECT "table1"."col1", "table1"."col2", "table1"."col3" FROM table1 ',
            'ORDER BY "table1"."col1" ASC'
        ])
        self.assertEqual(str(query), expected_text)

    @property
    def mock_tables(self):
        mt = {}
        for table in [self.mock_table1, self.mock_table2]:
            mt[table._name] = table
        return mt


class TestQueryFilters(NoofaTest):
    """
    Тестирование классов фильтров запросов.
    """
    @classmethod
    def setUpClass(cls):
        cls.wildcard = '%s'
        cls.col = 'col'
        cls.value = 10
        cls.comparison_template = f'{cls.col} %s %{cls.wildcard}'
        cls.args = (
            cls.col,
            cls.value,
            cls.wildcard
        )

    def test_eq(self):
        f = EqFilter(*self.args)
        self.assertEqual(str(f), self.comparison_template % '=')

    def test_neq(self):
        f = NeqFilter(*self.args)
        self.assertEqual(str(f), self.comparison_template % '<>')

    def test_gt(self):
        f = GtFilter(*self.args)
        self.assertEqual(str(f), self.comparison_template % '>')

    def test_ge(self):
        f = GeFilter(*self.args)
        self.assertEqual(str(f), self.comparison_template % '>=')

    def test_lt(self):
        f = LtFilter(*self.args)
        self.assertEqual(str(f), self.comparison_template % '<')

    def test_le(self):
        f = LeFilter(*self.args)
        self.assertEqual(str(f), self.comparison_template % '<=')

    def test_contains(self):
        f = ContainsFilter(*self.args)
        self.assertEqual(str(f), f'{self.col} LIKE {self.wildcard}')
        self.assertEqual(f._params, [f'%{self.value}%'])

    def test_startswith(self):
        f = StartsWithFilter(*self.args)
        self.assertEqual(str(f), f'{self.col} LIKE {self.wildcard}')
        self.assertEqual(f._params, [f'{self.value}%'])

    def test_endswith(self):
        f = EndsWithFilter(*self.args)
        self.assertEqual(str(f), f'{self.col} LIKE {self.wildcard}')
        self.assertEqual(f._params, [f'%{self.value}'])

    def test_in(self):
        f = InFilter(self.col, [1, 2], self.wildcard)
        self.assertEqual(str(f), 'col IN (%s, %s)')
        self._params = [1, 2]

    def test_not_in(self):
        f = NotInFilter(self.col, [1, 2], self.wildcard)
        self.assertEqual(str(f), 'col NOT IN (%s, %s)')
        self._params = [1, 2]



"""
Функции подключения к источникам.
"""

from ..sources.conn import get_source_class
from .base import ConnectionFunc, MandatoryArg


class NewConnection(ConnectionFunc):
    """
    Функция создания подключения к источнику.
    """
    description = 'Функция создания подключения к источнику'
    args_description = [
        MandatoryArg('Тип источника', 0, [str]),
        MandatoryArg('Строка подключения', 1, [str]),
    ]

    @classmethod
    def get_name(cls):
        return 'create_connection'

    def _operation(self, *args):
        conn = get_source_class(args[0])
        conn = conn(conn_str=args[1])
        return conn



import pathlib
import pandas
from abc import ABC, abstractmethod


class FileSource(ABC):
    """
    Абстрактный источник из файла.
    """
    source_type = ''

    def __init__(self, path, file_id=None, read_params={}):
        self.path = path
        self._read_params = read_params

    @abstractmethod
    def get_data(self):
        pass

    def test(self):
        path = pathlib.Path(self.path)
        return path.is_file()

    @property
    def is_sql(self):
        return False

    @property
    def source_type(self):
        return self.__class__.source_type

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        pass


class CSVSource(FileSource):
    """
    Источник из csv-файла.
    """
    source_type = 'csv'

    def get_data(self):
        self._read_params['sep'] = ';'
        data = pandas.read_csv(self.path, **self._read_params)
        return data


class ExcelSource(FileSource):
    """
    Источник из excel-файла.
    """
    source_type = 'excel'

    def get_data(self):
        data = pandas.read_excel(self.path, **self._read_params)
        return data


FILE_SOURCES = {
    'csv': CSVSource,
    'excel': ExcelSource,
}



from pandas import DataFrame, Series

from noofa.core.func import df as DF
from noofa.core.func.errors import (
    NotEnoughArguments,
    ArgumentTypeError,
)
from noofa.tests.base import NoofaTest


class TestDfFunctions(NoofaTest):
    """
    Тестирование функций работы с датафреймами.
    """
    @classmethod
    def setUpClass(cls):
        cls.df1 = DataFrame([
            {'a': 1, 'b': 'Hello', 'c': 3.14},
            {'a': 2, 'b': 'hell', 'c': 3.14},
            {'a': 2, 'b': 'ello', 'c': 3.14},
            {'a': 3, 'b': 'ohell', 'c': 3.14},
            {'a': 3, 'b': 'hellno', 'c': 3.14},
            {'a': 3, 'b': 'holle', 'c': 3.14},
        ])
        cls.df2 = DataFrame([
            {'a': 3, 'b': 'Hello', 'c': 3.14},
            {'a': 2, 'b': 'hell', 'c': 3.14},
            {'a': 2, 'b': 'ello', 'c': 3.14},
            {'a': 3, 'b': 'ohell', 'c': 3.14},
            {'a': 4, 'b': 'hellno', 'c': 3.14},
            {'a': 4, 'b': 'holle', 'c': 3.14},
        ])

    def test_join(self):
        f = DF.Join(self.df1, self.df2, 'a', 'a', 'inner')
        df = f()
        self.assertEqual(df.shape[0], 10)

    def test_join_raises_nea(self):
        f = DF.Join(self.df1, self.df2, 'a', 'a')
        self.assertRaises(NotEnoughArguments, f)

    def test_join_raises_ate(self):
        f = DF.Join(0, self.df2, 'a', 'a', 'inner')
        self.assertRaises(ArgumentTypeError, f)

        f = DF.Join(self.df1, 0, 'a', 'a', 'inner')
        self.assertRaises(ArgumentTypeError, f)

        f = DF.Join(self.df1, self.df2, 0, 'a', 'inner')
        self.assertRaises(ArgumentTypeError, f)

        f = DF.Join(self.df1, self.df2, 'a', 0, 'inner')
        self.assertRaises(ArgumentTypeError, f)

        f = DF.Join(self.df1, self.df2, 'a', 'a', 0)
        self.assertRaises(ArgumentTypeError, f)

    def test_union(self):
        f = DF.Union(self.df1, self.df2)
        df = f()
        self.assertEqual(df.shape[0], 12)

    def test_union_raises_nea(self):
        f = DF.Union(self.df1)
        self.assertRaises(NotEnoughArguments, f)

    def test_union_raises_ate(self):
        f = DF.Union(0, self.df2)
        self.assertRaises(ArgumentTypeError, f)

        f = DF.Union(self.df1, 0)
        self.assertRaises(ArgumentTypeError, f)

    def test_df_filter(self):
        f = DF.DfFilter('a', '>', 1)
        fd = f()
        self.assertIsInstance(fd, DF.DfFilterDict)
        self.assertCountEqual(list(fd._q.keys()), ['is_q', 'col_name', 'op', 'value'])

    def test_df_filter_raises_nea(self):
        f = DF.DfFilter('a', '>')
        self.assertRaises(NotEnoughArguments, f)

    def test_df_filter_raises_ate(self):
        f = DF.DfFilter(0, '>', 1)
        self.assertRaises(ArgumentTypeError, f)

        f = DF.DfFilter('a', 0, 1)
        self.assertRaises(ArgumentTypeError, f)

    def test_filter(self):
        f = DF.Filter(self.df1, DF.DfFilter('a', '>', 1)())
        df = f()
        self.assertEqual(df.shape[0], 5)

    def test_filter_raises_nea(self):
        f = DF.Filter(self.df1)
        self.assertRaises(NotEnoughArguments, f)

    def test_filter_raises_ate(self):
        f = DF.Filter(0, DF.DfFilter('a', '>', 1))
        self.assertRaises(ArgumentTypeError, f)

        f = DF.Filter(self.df1, 0)
        self.assertRaises(ArgumentTypeError, f)

    def test_add_column(self):
        f = DF.AddColumn(self.df1, 'd', 0)
        df = f()
        self.assertIn('d', df)

    def test_add_column_raises_nea(self):
        f = DF.AddColumn(self.df1, 'd')
        self.assertRaises(NotEnoughArguments, f)

    def test_add_column_raises_ate(self):
        f = DF.AddColumn(0, 'd', 0)
        self.assertRaises(ArgumentTypeError, f)

        f = DF.AddColumn(self.df1, 0, 0)
        self.assertRaises(ArgumentTypeError, f)

    def test_head(self):
        f = DF.Head(self.df1, 3)
        df = f()
        self.assertEqual(df.shape[0], 3)

    def test_head_raises_nea(self):
        f = DF.Head(self.df1)
        self.assertRaises(NotEnoughArguments, f)

    def test_head_raises_ate(self):
        f = DF.Head(0, 1)
        self.assertRaises(ArgumentTypeError, f)

        f = DF.Head(self.df1, '1')
        self.assertRaises(ArgumentTypeError, f)

    def test_tail(self):
        f = DF.Tail(self.df1, 3)
        df = f()
        self.assertEqual(df.shape[0], 3)

    def test_tail_raises_nea(self):
        f = DF.Tail(self.df1)
        self.assertRaises(NotEnoughArguments, f)

    def test_tail_raises_ate(self):
        f = DF.Tail(0, 1)
        self.assertRaises(ArgumentTypeError, f)

        f = DF.Tail(self.df1, '1')
        self.assertRaises(ArgumentTypeError, f)

    def test_count(self):
        f = DF.DfCount(self.df1)
        self.assertEqual(f(), 6)

    def test_count_raises_nea(self):
        f = DF.DfCount()
        self.assertRaises(NotEnoughArguments, f)

    def test_count_raises_ate(self):
        f = DF.DfCount(0)
        self.assertRaises(ArgumentTypeError, f)

        f = DF.Tail(self.df1, '1')
        self.assertRaises(ArgumentTypeError, f)

    def test_get_column(self):
        f = DF.GetColumn(self.df1, 'a')
        self.assertIsInstance(f(), Series)

    def test_get_column_raises_nea(self):
        f = DF.GetColumn(self.df1)
        self.assertRaises(NotEnoughArguments, f)

    def test_get_column_raises_ate(self):
        f = DF.GetColumn(0, 1)
        self.assertRaises(ArgumentTypeError, f)

        f = DF.GetColumn(self.df1, 0)
        self.assertRaises(ArgumentTypeError, f)




import math
from pandas import DataFrame
from datetime import datetime

from noofa.tests.base import NoofaFunctionTest
from noofa.core.func import numbers as N


class TestMathFunctions(NoofaFunctionTest):
    """
    Тестирование функций в выражениях.
    """
    @classmethod
    def setUpClass(cls):
        cls.add_cases_from_list(_CASES)


_df = DataFrame()
_dt = datetime.now()
_bad_args = ('1', _dt, _df)


"""
Содержимое кортежей в списке _CASES:
(
    Функция,
    сабкейсы (кортеж из аргументов и ожидаемого результата),
    аргументы для проверки ошибки по типу аргумента,
    аргументы для проверки ошибки по кол-ву аргументов
)
"""
_CASES = [
    (N.Abs, ((-1, 1), (1, 1)), _bad_args),
    (N.Acos, ((1, math.acos(1)), ), _bad_args),
    (N.Asin, ((1, math.asin(1)), ), _bad_args),
    (N.Atan, ((1, math.atan(1)), ), _bad_args),
    (N.Ceil, ((3.14, 4), ), _bad_args),
    (N.Cos, ((1, math.cos(1)), ), _bad_args),
    (N.Cot, ((1, 1/math.tan(1)), ), _bad_args),
    (N.Degrees, ((1, math.degrees(1)), ), _bad_args),
    (N.Div, (((10, 4), 2.5), ), (('1', 1), (_dt, 1), (1, _df)), (1, )),
    (N.Idiv, (((10, 4), 2), ), (('1', 1), (_dt, 1), (1, _df)), (1, )),
    (N.Exp, ((2, math.exp(2)), ), _bad_args),
    (N.Floor, ((3.14, 3), ), _bad_args),
    (N.Ln, ((3, math.log(3)), ), _bad_args),
    (N.Pow, (((3, 3), 27), ), (('1', 1), (_dt, 1), (1, _df)), (1, )),
    (N.Radians, ((1, math.radians(1)), ), _bad_args),
    (N.Round, ((3.14, 3), ((3.1415926, 2), 3.14)), (('1', 1), (_dt, 1), (1, _df))),
    (N.Sin, ((1, math.sin(1)), ), _bad_args),
    (N.Sqrt, ((9, 3), ), _bad_args),
    (N.Sum, ((1, 1), ([1, 2, 3], 6)), (), ()),
    (N.Tan, ((1, math.tan(1)), ), _bad_args),
]



from ..core.dataframes.panda_builder import pivot_table, empty
from .base import ReportComponent
from ..utils import get_dtypes
from .utils import apply_filters


class ReportTable(ReportComponent):
    """
    Компонент-таблица.
    """
    def __init__(self, exclude=None, aliases=None, **options):
        super().__init__(**options)
        self.type = 'table'
        self._to_exclude = exclude or []  # столбцы, которые исключ. при выводе

        # словарь для переименования выводимых столбцов;
        # должен иметь формат {название_существующего_столбца1: новое_название, ...}
        self._aliases = aliases or {}
        self._df = None  #  датафрейм, данные из которого будут выводиться в виде таблицы

    @property
    def df(self):
        return self._df

    @df.setter
    def df(self, value):
        self._df = value

    def build(self, **kwargs):
        """
        Построение таблицы. Под этим подразумевается переименование столбцов
        и отбрасывание лишних столбцов в зависимости от конфигурации таблицы.
        """
        self.df = self.evaluator.evaluate(self.base)

        if self._df is not None:
            if self._aliases:
                self.df = self.df.rename(columns=self._aliases)

            cols = self._to_exclude
            if cols:
                if len(cols) < len(self.df.columns):
                    self.df = self.df.drop(cols, axis=1)
        return self

    def to_csv(self, path=None):
        sep, idx = ';', False
        if path is None:
            return self.df.to_csv(sep=sep, index=idx)
        file = open(path, 'w')
        return self.df.to_csv(file, sep=sep, index=idx)

    def to_excel(self, path):
        return self.df.to_excel(path) if self.df is not None else None

    @property
    def header(self):
        """
        Список наименований столбцов таблицы.
        """
        return self.df.columns.to_list()

    @property
    def body(self):
        """
        Список списков значений соотв. строк таблицы.
        """
        recs = list(self.df.to_records(index=False))
        return [list(r) for r in recs]

    @property
    def data(self):
        """
        Список списков значений ячеек таблицы, включая заголовок.
        """
        data = [self.header]
        for r in self.body:
            data.append(r)
        return data

    @property
    def raw_data(self):
        return self.df.to_dict(orient='records')

    def fillna(self, fill_value=''):
        return self.df.fillna(fill_value)

    def to_dict(self):
        return {
            'header': self.header,
            'body': self.body,
            'title': self.title_text,
            'id': self.id,
        }


class ReportPivotTable(ReportComponent):
    """
    Сводная таблица.
    """
    def __init__(self, pivot_conf=None, **options):
        super().__init__(**options)
        self.type = 'pivot_table'
        self._df = None
        self._pivot_conf = pivot_conf
        self.pivot_df = None

    @property
    def df(self):
        return self._df

    @df.setter
    def df(self, value):
        self._df = value

    @property
    def filters_list(self):
        return self._pivot_conf.get('filters', [])

    def build(self, filters=None, **kwargs):
        if not self._check_pivot_conf():
            self.pivot_df = empty()
            return self

        self.df = self.evaluator.evaluate(self.base)
        if self.df is not None and self._pivot_conf is not None:
            if filters:
                df = apply_filters(self.df, filters)
            else:
                df = self.df
            self.pivot_df = pivot_table(df, **self._pivot_conf)
        return self

    @property
    def is_pivot_table(self):
        return True

    def to_html(self, sparsify=True, classes=None, bold_rows=False, **kwargs):
        return self.pivot_df.to_html(
            sparsify=sparsify,
            bold_rows=bold_rows,
            classes=classes,
            **kwargs,
        )

    def _check_pivot_conf(self):
        pc = self._pivot_conf
        index, columns = pc.get('index', []), pc.get('columns', [])
        has_values = bool(pc.get('aggfunc', {}))
        has_groups = bool(index) or bool(columns)
        return has_values and has_groups



import math
from pandas import isna

from plotly.subplots import make_subplots
import plotly.graph_objects as go
import plotly.express as px


def pivot_chart(df, chart_type='line', chart_params={}, **kwargs):
    if chart_type == 'line':
        return _line(df, **chart_params)
    if chart_type == 'pie':
        return _pie(df, **chart_params)
    if chart_type == 'bar':
        return _bar(df, **chart_params)
    if chart_type == 'scatter':
        return _scatter(df, **chart_params)


def _line(df, groupby=[], values=[], x='', y='', misc={}, **kwargs):
    vtp = {
        'method_kw': {
            'mode': 'lines' if not misc.get('markers', False) else 'lines+markers',
            'line_shape': 'linear' if misc.get('line_shape', 'linear') == 'linear' else 'spline',
        },
        'method': go.Scatter,
    }

    return _cartesian_plot(
        df, groupby=groupby, values=values,
        x=x, y=y, method=px.line,
        value_trace_params=vtp,
        plot_params=misc,
    )


def _bar(df, groupby=[], values=[],
    x='', y='', misc={}, **kwargs):

    vtp = {
        'method_kw': {
            'orientation': misc.get('orientation', 'v'),
            'textposition': 'auto',
            'bar_text': misc.get('text_auto', False),
            'text': '',
        },
        'method': go.Bar,
    }

    return _cartesian_plot(
        df, groupby=groupby, values=values,
        x=x, y=y, method=px.bar,
        value_trace_params=vtp,
        plot_params=misc,
    )


def _scatter(df, groupby=[], values=[],
    x='', y='', misc={}, **kwargs):

    vtp = {
        'method_kw': {
            'orientation': misc.get('orientation', 'v'),
            'mode': 'markers',
        },
        'method': go.Scatter,
    }

    return _cartesian_plot(
        df, groupby=groupby, values=values,
        x=x, y=y, method=px.scatter,
        value_trace_params=vtp,
        plot_params=misc,
    )


def _pie(df, groupby=[], values=[],
    x='', y='', misc={}, **kwargs):
    pie = px.pie
    exc = ('', '--')
    not_xy = x in exc or y in exc

    # если нет группировки
    if not groupby:
        if not_xy:
            return pie()
        return pie(df, values=y, names=x, **misc)

    # количество диаграмм равно количеству значений
    groups, subplots = len(groupby), len(values)

    # если есть группировка и 1 значение
    if groups and subplots == 1:
        value = values[0]
        df, indices_col, agg_col_name = _slice(df, groupby, value)
        return pie(df, names=indices_col, values=agg_col_name, **misc)

    # если есть группировка и значений несколько
    if groups and subplots > 1:
        sq = math.sqrt(subplots)
        n = math.ceil(sq)
        cols_n = n
        rows_n = math.ceil(subplots/n)
        spec = {"type": "domain"}
        specs = []
        i = 0
        s = []
        while i != cols_n*rows_n:
            s.append(spec)
            if len(s) == cols_n:
                specs.append(s)
                s = []
            i += 1
        st = [f'{v["func"]}({v["column"]})' for v in values]
        fig = make_subplots(
            cols=cols_n, rows=rows_n,
            specs=specs, subplot_titles=st)

        r, c = 1, 1
        for v in values:
            _, indices_col, agg_col_name = _slice(df, groupby, v)
            fig.add_trace(
                go.Pie(
                    name=f'{agg_col_name}',
                    labels=_[indices_col],
                    values=_[agg_col_name],
                    hovertemplate='%s=%%{label}<br>%s=%%{value}<extra></extra>' % (indices_col, agg_col_name),
                ),
                row=r, col=c
            )
            if c + 1 > cols_n:
                c = 1
                r += 1
            else:
                c += 1

        return fig

    return pie(**misc)


def _slice(df, groupby, value):
    col, func = value['column'], value['func']
    agg_df = df.groupby(groupby).agg({col: func})
    agg_df, indices_col = _append_indices(agg_df, groupby)
    agg_col_name = f'{func}({col})'
    agg_df[agg_col_name] = agg_df[col]
    return agg_df, indices_col, agg_col_name


def _append_indices(df, groupby):
    indices = list(df.index)
    if len(groupby) > 1:
        try:
            indices = [', '.join(i) for i in indices]
        except TypeError:
            _ = []
            for ind in indices:
                _.append(', '.join([str(i) for i in ind]))
            indices = _
    df = df.reset_index()
    indices_col = ', '.join(groupby)
    df[indices_col] = indices
    return df, indices_col


def _add_sline(fig, df, value, orientation='v', group=None):
    col, func = value['column'], value['func']
    ser = df.agg({col: func})
    v = float(ser[0])
    ann = f'{func}({col})={v}'
    if group:
        ann += f' ({group})'

    kw = {
        'line_dash': 'dot',
        'annotation_text': ann,
        'annotation_position': 'bottom left',
    }
    if orientation == 'h':
        fig.add_vline(x=v, **kw)
    else:
        fig.add_hline(y=v, **kw)
    return fig


def _add_groupby_col(df, groupby):
    groupby_col = ', '.join(groupby)
    for n, col in enumerate(groupby):
        if n == 0:
            df[groupby_col] = df[col].astype('str')
            continue
        df[groupby_col] += ', ' + df[col].astype('str')
    return df, groupby_col


def _get_value_trace(df, groupby, value, method=go.Scatter,
    orientation='v', method_kw={}, **kwargs):
    col, func = value['column'], value['func']
    x_values, y_values, names = [], [], []
    grouped = df.groupby(groupby)
    for name, group in grouped:
        y_value = group.agg({col: func})
        y_value = float(y_value[0])
        if not isna(y_value):
            x_values.append(name)
            y_values.append(y_value)
    label = f'{func}({col})'

    text = method_kw.setdefault('bar_text', False)
    if text:
        method_kw['text'] = y_values
    if orientation == 'h':
        x_values, y_values = y_values, x_values
        if text:
            method_kw['text'] = x_values
    method_kw.pop('bar_text')
    vl = method(
        name=label,
        x=x_values,
        y=y_values,
        hovertemplate='%s<br>%s=%%{x}<br>%%{y}<extra></extra>' % (label, groupby),
        **method_kw,
    )
    return vl


def _cartesian_plot(df, groupby=[], values=[], x='', y='',
    method=px.line, exc=['', '--'],
    plot_params={}, value_trace_params={}, **kwargs):

    plot = method
    empty = plot()
    not_xy = x in exc or y in exc

    orientation = plot_params.get('orientation', 'v')
    if orientation == 'h':
        x, y = y, x

    # нет группировки
    if not groupby:
        if not_xy:
            return empty

        fig = plot(df, x=x, y=y, **plot_params)
        if not values:
            return fig
        # если есть значения
        for value in values:
            fig = _add_sline(fig, df, value, orientation=orientation)
        return fig

    # если группировать по одному полю
    if len(groupby) == 1:
        groupby_col = groupby[0]
    else:
        df, groupby_col = _add_groupby_col(df, groupby)

    # если нет значений
    if not values:
        if not_xy:
            return empty
        return plot(df, x=x, y=y, color=groupby_col, **plot_params)

    # если не указаны x или y
    if not_xy:
        fig = plot(**plot_params)
        for value in values:
            fig.add_trace(
                _get_value_trace(df, groupby_col, value,
                    orientation=orientation, **value_trace_params)
            )
        return fig

    # если есть x, y, группировка и значения,
    # то для каждой линии достраиваются пунктирные линии значений
    fig = plot(df, x=x, y=y, color=groupby_col, **plot_params)
    grouped = df.groupby(groupby_col)
    for name, group in grouped:
        for value in values:
            fig = _add_sline(fig, group, value,
                group=name, orientation=orientation)
    return fig



import math
import statistics
from datetime import datetime, date, time
from pandas import DataFrame, Series

from noofa.tests.base import NoofaInterpreterTest


_df_expr = ''.join([
    'dataframe(',
    'list(',
    'list("a", "b", "c"),'
    'list(1,2,3),',
    'list(4,5,6)',
    ')',
    ')'
])


class TestFunctionsExpressions(NoofaInterpreterTest):
    """
    Тестирование функций интерпретатора в выражениях.
    """
    @classmethod
    def setUpClass(cls):
        cls.set_up_interpreter()
        cls.math_eq_cases = _MATH_EQ_CASES
        cls.logic_eq_cases = _LOGIC_EQ_CASES
        cls.strings_eq_cases = _STRINGS_EQ_CASES
        cls.date_eq_cases = _DATE_EQ_CASES
        cls.conv_eq_cases = _CONV_EQ_CASES
        cls.ds_eq_cases = _DS_EQ_CASES
        cls.df_eq_cases = _DF_EQ_CASES
        cls.stats_eq_cases = _STATS_EQ_CASES
        cls.df_expr = _df_expr

    def test_equality_cases(self):
        for case in self.equality_cases:
            expression, expected = case
            with self.subTest(expression):
                res = self.evaluate(expression)
                self.assertEqual(res, expected)

    @property
    def equality_cases(self):
        cases = ()
        for case in [
            self.math_eq_cases,
            self.logic_eq_cases,
            self.strings_eq_cases,
            self.date_eq_cases,
            self.conv_eq_cases,
            self.ds_eq_cases,
            self.df_eq_cases,
        ]:
            cases += case
        return cases

    def test_dataframe(self):
        df = self.evaluate(self.df_expr)
        self.assertIsInstance(df, DataFrame)
        self.assertEqual(df.shape, (2, 3))

    def test_date(self):
        d1 = self.evaluate('date()')
        d2 = datetime.now().date()
        self.assertEqual(d1, d2)

    def test_time(self):
        t = self.evaluate('time()')
        self.assertIsInstance(t, time)

    def test_now(self):
        n = self.evaluate('now()')
        self.assertIsInstance(n, datetime)


"""
Кейсы для проверки равенства результата ввыражения и ожидаемого значения.
Первый элемент в кортежах - выражение, второй - ожидаемый результат.
"""
_MATH_EQ_CASES = (
    ('abs(-1)', 1),
    ('abs(1)', 1),
    ('acos(1)', math.acos(1)),
    ('asin(1)', math.asin(1)),
    ('atan(1)', math.atan(1)),
    ('ceil(3.14)', 4),
    ('floor(3.54)', 3),
    ('cos(1)', math.cos(1)),
    ('cot(1)', 1/math.tan(1)),
    ('degrees(1)', math.degrees(1)),
    ('radians(1)', math.radians(1)),
    ('div(9, 2)', 4.5),
    ('idiv(9, 2)', 4),
    ('exp(3)', math.exp(3)),
    ('ln(3)', math.log(3)),
    ('log(3, 5)', math.log(3, 5)),
    ('pow(3, 3)', 27),
    ('round(3.14)', 3),
    ('round(3.54)', 4),
    ('round(3.1415926, 2)', 3.14),
    ('sin(1)', math.sin(1)),
    ('cos(1)', math.cos(1)),
    ('sqrt(13)', math.sqrt(13)),
    ('sum(1)', 1),
    ('sum(1, 2, 3)', 6),
    ('sum(list(1, 2, 3))', 6),
    ('tan(1)', math.tan(1)),
)


_LOGIC_EQ_CASES = (
    ('true()', True),
    ('false()', False),
    ('eq(1, 2)', False),
    ('eq(2, 2)', True),
    ('neq(1, 2)', True),
    ('neq(2, 2)', False),
    ('gt(2, 1)', True),
    ('gt(1, 2)', False),
    ('gte(2, 1)', True),
    ('gte(2, 2)', True),
    ('gte(1, 2)', False),
    ('lt(2, 1)', False),
    ('lt(1, 2)', True),
    ('lte(2, 1)', False),
    ('lte(2, 2)', True),
    ('lte(1, 2)', True),
    ('if(true(), 1, 2)', 1),
    ('if(false(), 1, 2)', 2),
    ('and(true(), false())', False),
    ('and(true(), true())', True),
    ('or(true(), false())', True),
    ('or(true(), true())', True),
    ('or(false(), false())', False),
    ('in(1, list())', False),
    ('in(1, list(1, 2, 3))', True),
    ('switch(1, 0, list(1, 10), list(2, 20))', 10),
    ('switch(2, 0, list(1, 10), list(2, 20))', 20),
    ('switch(3, 0, list(1, 10), list(2, 20))', 0),
    ('switch(1, 0)', 0),
)


_STRINGS_EQ_CASES = (
    ('contains("12", "1")', True),
    ('contains("12", "3")', False),
    ('endswith("abc", "bc")', True),
    ('endswith("abc", "bce")', False),
    ('startswith("abc", "ab")', True),
    ('startswith("abc", "bc")', False),
    ('len("abc")', 3),
    ('lower("ABC")', 'abc'),
    ('upper("abc")', 'ABC'),
    ('concat("a", "b")', 'ab'),
    ('concat("a", "b", "c", "d")', 'abcd'),
    ('join(",", "a", "b")', "a,b"),
    ('join(", ", "a", "b", "c")', "a, b, c"),
)


_DATE_EQ_CASES = (
    ('sectodate(12345)', datetime(1970, 1, 1, 3, 25, 45)),
    ('msectodate(12345000)', datetime(1970, 1, 1, 3, 25, 45)),
    ('sectodatestr(12345)', '1970-01-01 03:25:45'),
    ('sectodatestr(12345, "%H:%M:%S %d-%m-%Y")', '03:25:45 01-01-1970'),
    ('msectodatestr(12345000)', '1970-01-01 03:25:45'),
    ('msectodatestr(12345000, "%H:%M:%S %d-%m-%Y")', '03:25:45 01-01-1970'),
    ('datetosec(sectodate(12345))', 12345),
    ('datetostr(sectodate(12345))', '1970-01-01 03:25:45'),
    ('datetostr(sectodate(12345), "%H:%M:%S %d-%m-%Y")', '03:25:45 01-01-1970'),
    ('strtodate("1970-01-01 03:25:45")', datetime(1970, 1, 1, 3, 25, 45)),
    ('strtodate("03:25:45 01-01-1970", "%H:%M:%S %d-%m-%Y")', datetime(1970, 1, 1, 3, 25, 45)),
)


_CONV_EQ_CASES = (
    ('to_str(5)', '5'),
    ('to_str(3.14)', '3.14'),
    ('to_str("a")', 'a'),
    ('to_int(3.14)', 3),
    ('to_int(3)', 3),
    ('to_int("3")', 3),
    ('to_float(3.14)', 3.14),
    ('to_float(3)', 3.0),
    ('to_float("3.14")', 3.14),
    ('to_float(".5")', 0.5),
)


_DS_EQ_CASES = (
    ('list()', []),
    ('list(1, 2, 3)', [1, 2, 3]),
    (f'to_list(get_column({_df_expr}, "a"))', [1, 4]),
)


_DF_EQ_CASES = (
    (f'df_count({_df_expr})', 2),
    (f'df_count(df_head({_df_expr}, 1))', 1),
    (f'df_count(df_tail({_df_expr}, 1))', 1),
    (f'df_count(df_union({_df_expr}, {_df_expr}))', 4),
    (f'df_count(df_join({_df_expr}, {_df_expr}, "a", "a", "inner"))', 2),
    (''.join([
        'df_count(',
        f'filter({_df_expr}, df_filter("a", "==", 1))'
        ')'
    ]), 1),
    (''.join([
        'df_count(',
        'filter(',
        f'add_column({_df_expr}, "d", list(7, 8)),'
        'df_filter("d", "==", 7)',
        ')',
        ')'
    ]), 1),
)


_ = [1, 2, 3, 4, 5]
_l = 'list(1, 2, 3, 4, 5)'
_STATS_EQ_CASES = (
    ('mean(list(1, 2, 3, 4))', 2.5),
    ('min(list(1, 2, 3, 0, 5))', 0),
    ('max(list(1, 2, 3, 0, 5))', 5),
    ('mode(list(1, 2, 1, 2, 1))', 1),
    (f'median({_l})', statistics.median(_)),
    (f'stdev({_l})', statistics.stdev(_, 1)),
    (f'stdev({_l}, 0)', statistics.stdev(_, 0)),
    (f'variance({_l})', statistics.variance(_, 1)),
    (f'variance({_l}, 0)', statistics.variance(_, 0)),
)



from .core.func.errors import (
    ExpressionParsingError,
    ExpressionSyntaxError,
    ExpressionEvaluationError,
    NotEnoughArguments,
    InterpreterContextError,
)

from .components.exceptions import SchemaComponentNotFound
from .builders.exceptions import RecursiveDataframeBuildError


_COMPONENTS_DESC = {
    'source': ['источник', 'm'],
    'query': ['запрос', 'm'],
    'df': ['датафрейм', 'm'],
    'table': ['таблица', 'f'],
    'figure': ['график', 'm'],
}


class SchemaComponentNotFound(Exception):
    def __init__(self, component_id, component_type=None):
        self.component_id = component_id
        self.component_type = component_type

    def __str__(self):
        comp_id = self.component_id
        component_desc = _COMPONENTS_DESC.get(self.component_type, ['компонент', 'm'])
        component_type, gender = component_desc
        found = 'найдена' if gender == 'f' else 'найден'
        return f'В схеме данных не {found} {component_type} "{comp_id}"'


"""
Фильтры для датафреймов.
"""

from abc import ABC, abstractmethod


class PandaFilter(ABC):
    """
    Простой фильтр для датафреймов.
    """
    def __init__(self, panda_col, value):
        self._panda_col = panda_col  # название столбца
        self._value = value  # значение, с которым сравнивается значение в ячейке
        self._df = None

    def __and__(self, panda_filter):
        return self.filter & panda_filter.filter

    def __or__(self, panda_filter):
        return self.filter | panda_filter.filter

    @property
    def df(self):
        return self._df

    @df.setter
    def df(self, df):
        self._df = df

    @property
    @abstractmethod
    def filter(self):
        pass


class PandaQ:
    """
    Составной фильтр для датафреймов.
    """
    def __init__(self, *panda_filters):
        self._filter = None
        for f in panda_filters:
            if self.is_empty:
                self._filter = f.filter
            else:
                self._filter = (self._filter) & f.filter

    @property
    def is_empty(self):
        return self._filter is None

    def __and__(self, panda_q):
        """
        Логическое умножение с другим составным фильтром.
        """
        if type(panda_q) is PandaQ:
            f = panda_q.filter
            self_f = self.filter
            if f is not None:
                if self_f is not None: 
                    self._filter = self.filter & f
                    return self
                else:
                    return panda_q
        return self

    def __or__(self, panda_q):
        """
        Логическое сложение с другим составным фильтром.
        """
        if type(panda_q) is PandaQ:
            f = panda_q.filter
            self_f = self.filter
            if f is not None:
                if self_f is not None:
                    self._filter = self.filter | f
                    return self
                else:
                    return panda_q
        return self

    @property
    def filter(self):
        return self._filter


class PandaEq(PandaFilter):
    """
    Простой фильтр равенства.
    """
    @property
    def filter(self):
        return self._df[self._panda_col] == self._value


class PandaNeq(PandaFilter):
    """
    Простой фильтр неравенства.
    """
    @property
    def filter(self):
        return self._df[self._panda_col] != self._value


class PandaGte(PandaFilter):
    """
    Простой фильтр 'больше либо равно'.
    """
    @property
    def filter(self):
        return self._df[self._panda_col] >= self._value


class PandaGt(PandaFilter):
    """
    Простой фильтр 'больше, чем'.
    """
    @property
    def filter(self):
        return self._df[self._panda_col] > self._value


class PandaLte(PandaFilter):
    """
    Простой фильтр 'меньше либо равно'.
    """
    @property
    def filter(self):
        return self._df[self._panda_col] <= self._value


class PandaLt(PandaFilter):
    """
    Простой фильтр 'меньше, чем'.
    """
    @property
    def filter(self):
        return self._df[self._panda_col] < self._value


class PandaContains(PandaFilter):
    """
    Простой фильтр проверки наличия значения.
    """
    @property
    def filter(self):
        return self._df[self._panda_col].str.contains(self._value)


class PandaStartsWith(PandaFilter):
    """
    Простой фильтр проверки того, начинается ли строка с подстроки.
    """
    @property
    def filter(self):
        return self._df[self._panda_col].str.lower().str.startswith(self._value)


class PandaEndsWith(PandaFilter):
    """
    Простой фильтр проверки того, заканчивается ли строка подстрокой.
    """
    @property
    def filter(self):
        return self._df[self._panda_col].str.lower().str.endswith(self._value)


class PandaIn(PandaFilter):
    """
    Простой фильтр проверки включенности элемента в список.
    """
    @property
    def filter(self):
        return self._df[self._panda_col].isin(self._value)


PANDA_FILTERS = {
    '>': PandaGt,
    '>=': PandaGte,
    '<': PandaLt,
    '<=': PandaLte,
    '==': PandaEq,
    '!=': PandaNeq,
    'contains': PandaContains,
    'startswith': PandaStartsWith,
    'endswith': PandaEndsWith,
    'in': PandaIn,
}


def _parse_filter(df, panda_jsf):
    if panda_jsf['is_q'] == True:
        filters = []
        for f in panda_jsf['filters']:
            filters.append(_parse_filter(df, f))
        op = panda_jsf['op']
        if op == 'or':
            panda_filter = PandaQ()
            for f in filters:
                panda_filter |= f
        else:
            panda_filter = PandaQ()
            for f in filters:
                panda_filter &= f
        return panda_filter
    else:
        col, op, value = panda_jsf['col_name'], panda_jsf['op'], panda_jsf['value']
        panda_filter = PANDA_FILTERS[op](col, value)
        panda_filter.df = df
        return PandaQ(panda_filter)

def _parse_filters(df, panda_jsfilters):
    """
    Построение фильтра из json.
    """
    res = PandaQ()
    for f in panda_jsfilters:
        p = _parse_filter(df, f)
        res &= p
    return res


"""
Для работы с датафреймами.
"""
from datetime import datetime, date, time
import pandas as pd

from . import filters


def new(data={}, columns=[]):
    return pd.DataFrame(data, columns=columns)


def join(df1, df2, on, how='inner'):
    """
    Соединение датафреймов.
    """
    result_df = pd.merge(
        df1, df2,
        left_on=on[0], right_on=on[1],
        how=how,
    )
    return result_df


def union(dataframes_list):
    """
    Склеивание датафреймов из списка.
    dataframes_list - список датафреймов pandas.
    """
    result = dataframes_list.pop(0)
    for df in dataframes_list:
        result = pd.concat([result, df], ignore_index=True)
    return result


def empty():
    """
    Пустой датафрейм.
    """
    return pd.DataFrame()


def add_column(df, col_name, col_data):
    """
    Добавление столбца к датафрейму.
    """
    #df[col_name] = col_data
    new_df = df.assign(**{col_name: col_data})
    return new_df


def add_columns(df, columns):
    return df.assign(**columns)


def rename_columns(df, aliases_dict, inplace=True):
    """
    Переименование столбца датафрейма.
    df - датафрейм pandas,
    aliases_dict - словарь вида {название_поля:новое название}.
    """
    df.rename(columns=aliases_dict, inplace=inplace)
    return df


def drop_columns(df, cols_list, inplace=True):
    """
    Удаление столбцов.
    """
    for col in cols_list:
        df.drop(col, inplace=inplace, axis=1)
    return df


def filter(df, panda_jsfilters):
    """
    Фильтрация строк датафрейма.
    """
    panda_filter = filters._parse_filters(df, panda_jsfilters)
    return df[panda_filter.filter]


def order_by(df, by, asc=True):
    """
    Упорядочивание строк датафрейма.
    """
    return df.sort_values(by=by, ascending=asc)


def get_filter(filter_type):
    return filters.PANDA_FILTERS[filter_type]


def lazy_filter(df, pf):
    if isinstance(pf, filters.PandaQ):
        return df[pf.filter]
    return df[pf]


def astype(df, col=None, dtype=None):
    if dtype == 'datetime':
        dtype = 'datetime64[s]'

    dtype_is_float = dtype == 'float'
    current_types = {c: str(t) for c, t in df.dtypes.to_dict().items()}
    cur_type = current_types[col]
    dt_to_float = cur_type == 'datetime64[ns]' and dtype_is_float
    ns_to_num = cur_type == 'datetime64[ns]' and dtype in ['int', 'float']

    if dt_to_float or ns_to_num:
        df[col] = df[col].astype('int64')
    if ns_to_num:
        df[col] = df[col]/10**9

    return df.astype({col: dtype})


def pivot_table(df, index=None, columns=None, values=None,
    aggfunc='count', fill_value=None, margins=False, **kwargs):
    """
    Сводная таблица.
    """
    if not aggfunc or df.empty:
        return pd.DataFrame()

    agg_func = aggfunc
    if aggfunc and isinstance(aggfunc, dict):
        agg_func = {}
        for col, func in aggfunc.items():
            agg_func[col] = func if isinstance(func, list) and len(func) > 1 else func[0]

    values = list(agg_func.keys()) if margins == True else None
    res = df.pivot_table(index=index, columns=columns, values=values,
        aggfunc=agg_func, fill_value=fill_value, margins=margins)
    return res


def fill_na(df, col='', fill_value=None, **kwargs):
    """
    Заполнение пустых значений.
    """
    assert isinstance(
        fill_value,
        (bool, str, int, float, date, time, datetime),
    ), f'Нельзя использовать значения типа {type(fill_value)}'

    if not col:
        return df.fillna(fill_value) if fill_value is not None else df

    try:
        df[col] = df.loc[:, col].fillna(fill_value)
    except:
        pass
    return df


def drop_na(df, col):
    """
    Отбрасывание строк, в столбцах col которых
    находятся пустые значения.
    """
    try:
        df = df[df[col].notnull()]
    except:
        pass
    return df



from . import panda_builder


_ISNAN = panda_builder.pd.isna


def get_df_descriptor(df, ignore_nan_as_unique=False):
    """
    Получить описыватель датафрейма.
    """
    return DfDescriptor(df, ignore_nan_as_unique=ignore_nan_as_unique)


class DfDescriptor:
    """
    Описыватель датафрейма. Содержит список столбцов,
    уникальных значений и типов данных.

    ignore_nan_as_unique - учитывать ли значения nan при получении
    списка уникальных значений по столбцам. Не учитывать при True.
    """
    def __init__(self, df, ignore_nan_as_unique=False):
        self.columns = sorted(list(df.columns))
        self.dtypes = get_dtypes(df)
        self.unique = get_unique(df, ignore_nan_as_unique=ignore_nan_as_unique)

    def describe(self):
        return {
            'columns': self.columns,
            'dtypes': self.dtypes,
            'unique': self.unique,
        }


def get_dtypes(df):
    """
    Получение словаря с типами данных в столбцах дф.
    """
    _ = df.dtypes.to_dict()
    dtypes = {col: str(dtype) for col, dtype in _.items()}
    return dtypes


def get_unique(df, cols=None, ignore_nan_as_unique=False):
    """
    Получение уникальных значений по столбцу
    (при cols is None - по всем столбцам) датафрейма.
    """
    unique_values = {}
    if cols is None:
        cols = list(df.columns)
    for col in cols:
        u = df[col].unique()
        if ignore_nan_as_unique == True:
            u = [i for i in u if not _ISNAN(i)]
        else:
            u = list(u)

        try:
            u.sort()
        except:
            pass
        unique_values[col] = u

    return unique_values


def new_df(data, columns):
    return panda_builder.new(data, columns)



from pandas import DataFrame, Series
from datetime import datetime

from noofa.tests.base import NoofaFunctionTest
from noofa.core.func import strings as S


class TestStringsFunctions(NoofaFunctionTest):
    """
    Тестирование функций преобразования типов.
    """
    @classmethod
    def setUpClass(cls):
        cls.add_cases_from_list(_CASES)


_dt = datetime.now()
_bad_args = [_dt, _dt.date(), _dt.time(), DataFrame(), Series((1,2))]

"""
Содержимое кортежей в списке _CASES:
(
    Функция,
    сабкейсы (кортеж из аргументов и ожидаемого результата),
    аргументы для проверки ошибки по типу аргумента,
    аргументы для проверки ошибки по кол-ву аргументов
)
"""
_CASES = [
    (S.Contains,
        (
            (('12', '1'), True),
            (('12', '3'), False),
        ),
        (('1', 1), ('2', 3.14), ),
        (1, ),
    ),
    (S.Endswith,
        (
            (('12', '2'), True),
            (('12', '3'), False),
        ),
        (('1', 1), ('2', 3.14), ),
        (1, ),
    ),
    (S.Startswith,
        (
            (('12', '1'), True),
            (('12', '3'), False),
        ),
        (('1', 1), ('2', 3.14), ),
        (1, ),
    ),
    (S.Len,
        (
            ('0123', 4),
        ),
        _bad_args,
        (),
    ),
    (S.Lower,
        (
            ('TRUE', 'true'),
        ),
        _bad_args,
        (),
    ),
    (S.Upper,
        (
            ('true', 'TRUE'),
        ),
        _bad_args,
        (),
    ),
    (S.Concat,
        (
            (('Hello', ', World'), 'Hello, World'),
            (('Hello', ',', ' ', 'World'), 'Hello, World'),
        ),
        ((1, ''), (3.14, ''), (_dt, '')),
        ('', ),
    ),
    (S.Join,
        (
            ((',', '1', '2'), '1,2'),
            ((', ', '1', '2', '3'), '1, 2, 3'),
        ),
        (('1', 1, ''), ('2', 3.14, ''), ),
        (1, ),
    ),
]



"""
Математические функции.
"""

import math
from .base import MathFunc, NonMandatoryArg, MandatoryArg


class Abs(MathFunc):
    """
    Модуль числа.
    """
    description = 'Получение модуля числа'
    operation = abs
    args_description = [
        MandatoryArg('Число', 0, [int, float]),
    ]


class Acos(MathFunc):
    """
    Арккосинус числа.
    """
    description = 'Получение арккосинуса числа'
    operation = math.acos
    args_description = [
        MandatoryArg('Число', 0, [int, float]),
    ]


class Asin(MathFunc):
    """
    Арксинус числа.
    """
    description = 'Получение арксинуса числа'
    operation = math.asin
    args_description = [
        MandatoryArg('Число', 0, [int, float]),
    ]


class Atan(MathFunc):
    """
    Арктангенс числа.
    """
    description = 'Получение арктангенса числа'
    operation = math.atan
    args_description = [
        MandatoryArg('Число', 0, [int, float]),
    ]


""" class Atan2(MathFunc):
    '''
    Арктангенс частного двух чисел.
    '''
    description = 'Получение арктангенса x/y'
    operation = math.atan2
    args_description = [
        MandatoryArg('x', 0),
        MandatoryArg('y', 1),
    ] """


class Ceil(MathFunc):
    """
    Округление числа до ближайшего целого
    в большую сторону.
    """
    description = 'Округление числа в большую сторону'
    operation = math.ceil
    args_description = [
        MandatoryArg('Число', 0, [int, float]),
    ]


class Cos(MathFunc):
    """
    Косинус угла в радианах.
    """
    description = 'Получение косинуса угла. Значение угла - в радианах'
    operation = math.cos
    args_description = [
        MandatoryArg('Угол(рад)', 0, [int, float]),
    ]


class Cot(MathFunc):
    """
    Котангенс угла.
    """
    description = 'Получение котангенса угла. Значение угла - в радианах'
    args_description = [
        MandatoryArg('Угол(рад)', 0, [int, float]),
    ]

    def _operation(self, *args):
        try:
            return 1/math.tan(args[0])
        except ZeroDivisionError:
            return math.inf


class Degrees(MathFunc):
    """
    Перевод радиан в градусы.
    """
    description = 'Перевод радиан в градусы'
    operation = math.degrees
    args_description = [
        MandatoryArg('Радианы', 0, [int, float]),
    ]


class Div(MathFunc):
    """
    Деление.
    """
    description = 'Деление двух чисел x/y с остатком'
    args_description = [
        MandatoryArg('x', 0, [int, float]),
        MandatoryArg('y', 1, [int, float]),
    ]

    def _operation(self, *args):
        try:
            return args[0]/args[1]
        except ZeroDivisionError:
            return math.inf


class Idiv(MathFunc):
    """
    Целочисленное деление.
    """
    description = 'Целочисленное деление двух чисел x/y'
    args_description = [
        MandatoryArg('x', 0, [int, float]),
        MandatoryArg('y', 1, [int, float]),
    ]

    def _operation(self, *args):
        try:
            return args[0]//args[1]
        except ZeroDivisionError:
            return math.inf


class Exp(MathFunc):
    """
    Вычисление экспоненты.
    """
    description = 'Вычисление экспоненты'
    operation = math.exp
    args_description = [
        MandatoryArg('Степень', 0, [int, float]),
    ]


class Floor(MathFunc):
    """
    Округление до ближайшего целого 
    в меньшую сторону.
    """
    description = 'Округление числа в меньшую сторону'
    operation = math.floor
    args_description = [
        MandatoryArg('Число', 0, [int, float]),
    ]


class Ln(MathFunc):
    """
    Натуральный логарифм числа.
    """
    description = 'Натуральный логарифм числа'
    args_description = [
        MandatoryArg('Число', 0, [int, float]),
    ]

    def _operation(self, *args):
        return math.log(args[0])


class Log(MathFunc):
    """
    Вычисление логарифма x по основанию y.
    """
    description = 'Вычисление логарифма числа по основанию'
    operation = math.log
    args_description = [
        MandatoryArg('Число', 0, [int, float]),
        MandatoryArg('Основание', 1, [int, float]),
    ]

    def _operation(self, *args):
        return math.log(args[0], args[1])


class Pow(MathFunc):
    """
    Возведение x в степень y.
    """
    description = 'Возведение x в степень y'
    operation = pow
    args_description = [
        MandatoryArg('x', 0, [int, float]),
        MandatoryArg('y', 0, [int, float]),
    ]


class Radians(MathFunc):
    """
    Перевод градусов в радианы.
    """
    description = 'Перевод градусов в радианы'
    operation = math.radians
    args_description = [
        MandatoryArg('Градусы', 0, [int, float]),
    ]


class Round(MathFunc):
    """
    Округление числа до определённого кол-ва знаков после запятой.
    """
    description = 'Округление числа до определённого количества n знаков после запятой'
    operation = round
    args_description = [
        MandatoryArg('Число', 0, [int, float]),
        NonMandatoryArg('n', 1, [int, float]),
    ]


class Sin(MathFunc):
    """
    Синус угла в радианах.
    """
    description = 'Получение синуса угла. Значение угла - в радианах'
    operation = math.sin
    args_description = [
        MandatoryArg('Угол(рад)', 0, [int, float]),
    ]


class Sqrt(MathFunc):
    """
    Извлечение квадратного корня числа.
    """
    description = 'Извлечение квадратного корня числа'
    operation = math.sqrt
    args_description = [
        MandatoryArg('Число', 0, [int, float]),
    ]


class Sum(MathFunc):
    description = 'Сумма нескольких чисел'
    args_description = [
        MandatoryArg('Слагаемое1', 0),
        NonMandatoryArg('Слагаемое2', 1),
    ]

    def _operation(self, *args):
        result = 0
        arg = args[0]
        if isinstance(arg, list):
            return sum(arg)
        else:
            try:
                return float(arg.sum()[0])
            except AttributeError:
                pass
        for arg in args:
            result += arg
        return result


class Tan(MathFunc):
    """
    Тангенс угла в радианах.
    """
    description = 'Получение тангенса угла. Значение угла - в радианах'
    operation = math.tan
    args_description = [
        MandatoryArg('Угол(рад)', 0, [int, float]),
    ]



"""
Логические функции.
"""
from datetime import datetime

from .base import LogicFunc, NonMandatoryArg, MandatoryArg


class BoolTrue(LogicFunc):
    """
    Функция, возвращающая логическое True.
    """
    description = 'Функция, возвращающая True'

    @classmethod
    def get_name(cls):
        return 'true'

    def _operation(self, *args):
        return True


class BoolFalse(LogicFunc):
    """
    Функция, возвращающая логическое False.
    """
    description = 'Функция, возвращающая False'

    @classmethod
    def get_name(cls):
        return 'false'

    def _operation(self, *args):
        return False


class Eq(LogicFunc):
    """
    Функция проверки равенства.
    """
    description = 'Проверка равенства двух значений'
    args_description = [
        MandatoryArg('Значение1', 0, [str, int, float, bool, datetime, list]),
        MandatoryArg('Значение2', 1, [str, int, float, bool, datetime, list]),
    ]

    def _operation(self, *args):
        return args[0] == args[1]


class Neq(LogicFunc):
    """
    Функция проверки неравенства.
    """
    description = 'Проверка неравенства двух значений'
    args_description = [
        MandatoryArg('Значение1', 0, [str, int, float, bool, datetime, list]),
        MandatoryArg('Значение2', 1, [str, int, float, bool, datetime, list]),
    ]

    def _operation(self, *args):
        return args[0] != args[1]


class Gt(LogicFunc):
    """
    Функция проверки x > y.
    """
    description = 'Сравнение Значение1 > Значение2'
    args_description = [
        MandatoryArg('Значение1', 0, [int, float, bool, datetime]),
        MandatoryArg('Значение2', 1, [int, float, bool, datetime]),
    ]

    def _operation(self, *args):
        return args[0] > args[1]


class Gte(LogicFunc):
    """
    Функция проверки x >= y.
    """
    description = 'Сравнение Значение1 >= Значение2'
    args_description = [
        MandatoryArg('Значение1', 0, [int, float, bool, datetime]),
        MandatoryArg('Значение2', 1, [int, float, bool, datetime]),
    ]

    def _operation(self, *args):
        return args[0] >= args[1]


class Lt(LogicFunc):
    """
    Функция проверки x < y.
    """
    description = 'Сравнение Значение1 < Значение2'
    args_description = [
        MandatoryArg('Значение1', 0, [int, float, bool, datetime]),
        MandatoryArg('Значение2', 1, [int, float, bool, datetime]),
    ]

    def _operation(self, *args):
        return args[0] < args[1]


class Lte(LogicFunc):
    """
    Функция проверки x <= y.
    """
    description = 'Сравнение Значение1 <= Значение2'
    args_description = [
        MandatoryArg('Значение1', 0, [int, float, bool, datetime]),
        MandatoryArg('Значение2', 1, [int, float, bool, datetime]),
    ]

    def _operation(self, *args):
        return args[0] <= args[1]


class If(LogicFunc):
    """
    Функция условия.
    """
    description = 'Функция условия IF'
    args_description = [
        MandatoryArg('Условие', 0),
        MandatoryArg('Значение при true', 1),
        MandatoryArg('Значение при false', 2),
    ]

    def _operation(self, *args):
        if args[0] == True:
            return args[1]
        return args[2]


class And(LogicFunc):
    """
    Функция логического умножения AND.
    """
    _arguments_type = bool
    description = 'Функция логического умножения AND'
    args_description = [
        MandatoryArg('Выражение1', 0),
        MandatoryArg('Выражение2', 1),
    ]

    def _operation(self, *args):
        return args[0] & args[1]


class Or(LogicFunc):
    """
    Функция логического сложения OR.
    """
    _arguments_type = bool
    description = 'Функция логического сложения OR'
    args_description = [
        MandatoryArg('Выражение1', 0),
        MandatoryArg('Выражение2', 1),
    ]

    def _operation(self, *args):
        return args[0] | args[1]


class In(LogicFunc):
    """
    Функция проверки вхождения значения в последовательность.
    """
    description = 'Функция проверки вхождения значения в последовательность'
    args_description = [
        MandatoryArg('Значение', 0),
        MandatoryArg('Последовательность', 1, [str, list]),
    ]

    def _operation(self, *args):
        return args[0] in args[1]


class Switch(LogicFunc):
    """
    Функция сравнения с несколькими вариантами.
    """
    description = 'Функция сравнения с несколькими вариантами'
    args_description = [
        MandatoryArg('Значение', 0),
        MandatoryArg('Результат по умолчанию', 1),
        NonMandatoryArg('Список с парой значение-результат', 2),
    ]

    def _operation(self, *args):
        value, default = args[0], args[1]
        cases = args[2:]
        if not cases:
            return default
        cases = {c[0]:c[1] for c in cases}
        try:
            result = cases[value]
        except KeyError:
            result = default
        return result



"""
Классы для схемы данных отчёта.
"""
from pandas import DataFrame

from ..core import collect_tables, get_source_class, Qbuilder
from .exceptions import SchemaComponentNotFound


class DataSchema:
    """
    Схема получения данных профиля.

    Содержит в себе значения по источникам, запросам и датафреймам,
    которые используются при формировании отчёта в классе ReportBuilder.
    """
    def __init__(self, sources=None, queries=None, dataframes=None, **kwargs):
        self._sources = sources if sources is not None else {}
        self._queries = queries if queries is not None else {}
        self._dataframes = dataframes if dataframes is not None else {}
        self._dataframes_by_name = {}

    def add_source(self, id=None, name='', type='', connection=None, **options):
        """
        Добавление источника в схему.
        """
        conn = connection
        if conn is None:
            source_cls = get_source_class(type)
            conn = source_cls(**options)
        self._sources[id] = SchemaSource(id=id, connection=conn, type=type, name=name)
        return self

    def add_query(self, id=None, name='', source=None, query_src=None,
        build_from='expression', **options):
        """
        Добавление запроса в схему.
        """
        name = name
        source_id = source
        src = self._sources[source_id]
        query_src = query_src
        build_from = build_from
        self._queries[id] = SchemaQuery(
            id, src, query_src, name=name, build_from=build_from)
        return self

    def add_dataframe(self, base=None, **options):
        """
        Добавление датафрейма в схему.
        """
        if base['type'] == 'query':
            source = self._sources.get(base['source'], None)
            query_id = base['value']
            query = self._queries[query_id]
            base['source'] = source
            base['query'] = query
        if base['type'] == 'source':
            source = self._sources.get(base['source'], None)
            base['source'] = source
        schema_dataframe = SchemaDataframe(base=base, **options)
        self._dataframes[schema_dataframe.id] = schema_dataframe
        self._dataframes_by_name[schema_dataframe.name] = schema_dataframe
        return self

    def get_source(self, source_id):
        source = self._sources.get(source_id, None)
        if source is None:
            raise SchemaComponentNotFound(source_id, 'source')
        return source

    def get_query(self, query_id):
        q = self._queries.get(query_id, None)
        if q is None:
            raise SchemaComponentNotFound(query_id, 'query')
        return q

    def get_dataframe(self, df_id):
        df = self._dataframes.get(df_id, None)
        if df is None:
            df = self._dataframes_by_name.get(df_id, None)
            if df is None:
                raise SchemaComponentNotFound(df_id, 'df')
        return df


class SchemaSource:
    """
    Источник в схеме профиля.
    """
    def __init__(self, id=None, type='', name='', connection=None, **kwargs):
        self.type = type
        self.id = id
        self.name = name
        self.connection = connection

    @property
    def is_opened(self):
        """
        Открыт ли источник (есть ли подключение).
        """
        return self.connection.connection is not None

    def open(self):
        """
        Открыть источник (соединение).
        """
        if not self.is_opened:
            self.connection.open()

    def close(self):
        self.connection.close()

    def test(self):
        self.connection.test()

    def get_table(self, table_name):
        """
        Построение объекта таблицы по имени.
        """
        self.open()
        return self.connection.get_table(table_name)

    def get_table_multiple(self, tables_list):
        self.open()
        return self.connection.get_table_multiple(tables_list)

    @property
    def is_sql(self):
        return self.connection.is_sql

    def get_data(self, query=None, **kwargs):
        source = self.connection
        if query is not None:
            if self.is_sql:
                return source.get_data(query=query)
            else:
                source.source = query['base']
                return source.get_data()
        return source.get_data()

    @property
    def wildcard(self):
        return self.connection.wildcard

    def __enter__(self):
        self.open()
        return self.connection

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()


class SchemaQuery:
    """
    Запрос в схеме профиля.
    """
    def __init__(self, id_, source=None, query_src=None,
        name='', build_from='expression', **kwargs):
        self.id = id_
        self.name = name
        self.query_src = query_src
        self.query = None
        self._source = source
        self._compiled = None
        self.build_from = build_from

    def execute(self):
        with self._source as source:
            if source.is_sql:
                self._compiled = self._compile()
                data = source.get_data(query=self._compiled)
            else:
                source.source = self.query_src['base']
                data = source.get_data()
            return data

    def _compile(self):
        """
        Создание объекта запроса (SelectQuery из noofa.core.conn.query).
        """
        query = self.query
        tables_list = collect_tables(query)
        source = self._source
        if not source.is_opened:
            source.open()
        tables = source.get_table_multiple(tables_list)
        qb_args = [tables, query, source.wildcard, source.connection.source_type == 'mssql']
        qbuilder = Qbuilder(*qb_args)
        return qbuilder.parse_query()


class SchemaDataframe:
    """
    Датафрейм в схеме профиля.
    """
    def __init__(self, id='', name='', base=None, dtypes=None, unions=None, joins=None,
        filters=None, ordering=None, columns=None, fillna=None, **options):
        self.id = id
        self.name = name
        self._build_type = base['type']
        self._build_from = base['value']
        self._source = base.get('source', None)
        self._query = base.get('query', None)
        self.dtypes = dtypes or []
        self.unions = unions or []
        self.joins = joins or []
        self.filters = filters or []
        self.ordering = ordering
        self.cols = columns or []
        self.fillna = fillna or []

    def get_data(self):
        if self.build_type == 'source':
            return self._source.get_data()
        data = self._query.execute()
        return data

    def build(self):
        res = self.get_data()
        if self.build_type == 'source':
            return res
        data, columns = res.data, res.columns
        return DataFrame(data, columns=columns)

    @property
    def build_type(self):
        return self._build_type

    @property
    def build_from(self):
        return self._build_from



from datetime import datetime

from ..sources.utils import Qbuilder, collect_tables
from .base import SqlFunc, MandatoryArg, NonMandatoryArg


class SqlDict:
    """
    Обертка для select-запросов и их частей в виде словарей.
    Используется для упрощенного сложения запросов и фильтров
    в виде словарей.
    """
    def __init__(self, dict_):
        self._q = dict_

    @property
    def q_part(self):
        return self._q

    def keys(self):
        return self._q.keys()

    def values(self):
        return self._q.values()

    def __getitem__(self, key):
        return self._q[key]


class SqlWhereDict(SqlDict):
    """
    Обертка для фильтров where в виде словарей.
    """
    def __and__(self, value):
        return SqlWhereDict({
            'is_complex': True,
            'op': 'AND',
            'filters': [self.q_part, value.q_part],
        })

    def __or__(self, value):
        return SqlWhereDict({
            'is_complex': True,
            'op': 'OR',
            'filters': [self.q_part, value.q_part],
        })


class SqlExecute(SqlFunc):
    """
    Функция выполнения sql-запроса.
    """
    group = 'sql'
    description = 'Функция выполнения sql-запроса'
    args_description = [
        MandatoryArg('Объект соединения', 0),
        MandatoryArg('Объект запроса', 1),
    ]

    @classmethod
    def get_name(cls):
        return 'sql_execute'

    def _operation(self, *args):
        conn = args[0]
        q = args[1].q_part
        tables_list = collect_tables(q)
        conn.open()
        tables = conn.get_table_multiple(tables_list)
        qb = Qbuilder(tables, q)
        query = qb.parse_query()
        data = conn.get_data(query=query)
        conn.close()
        return data


class SqlSelect(SqlFunc):
    """
    Функция построения select-запроса.
    """
    description = 'Функция построения select-запроса'
    args_description = [
        MandatoryArg('Базовая таблица', 0, [str]),
        NonMandatoryArg('Список соединений', 1, [list, SqlDict]),
        NonMandatoryArg('Список фильтров', 2, [list, SqlWhereDict]),
        NonMandatoryArg('Список сортировок', 3, [list, SqlDict]),
        NonMandatoryArg('Список выбираемых полей', 4, [list]),
        NonMandatoryArg('Количество строк', 5, [int, str]),
    ]

    @classmethod
    def get_name(self):
        return 'sql_select'

    def _operation(self, *args):
        jsq = {
            'is_subquery': True,
            'tables': [args[0], ],
            'base': args[0],
            'joins': [],
            'filters': [],
            'values': [],
            'order_by': [],
        }
        try:
            joins = args[1]
            if isinstance(joins, list):
                for j in joins:
                    j = j.q_part
                    tables = [j['l'], j['r']]
                    for table in tables:
                        jsq['tables'].append(table)
                    jsq['joins'].append(j)
            else:
                j = joins.q_part
                jsq['joins'].append(j)
                jsq['tables'] += [j['l'], j['r']]
        except IndexError:
            pass
        try:
            filters_ = args[2]
            if isinstance(filters_, list):
                jsq['filters'] = [f.q_part for f in args[2]]
            else:
                jsq['filters'].append(filters_.q_part)
        except IndexError:
            pass
        try:
            orderings = args[3]
            if isinstance(orderings, list):
                jsq['order_by'] = [o.q_part for o in orderings]
            else:
                jsq['order_by'].append(orderings.q_part)
        except IndexError:
            pass
        try:
            for v in args[4]:
                alias = None
                if isinstance(v, list):
                    tf, alias = v[0].split('.'), v[1]
                else:
                    tf = v.split('.')
                value = {'table': tf[0], 'field': tf[1]}
                if alias is not None:
                    value['alias'] = alias
                jsq['values'].append(value)
        except IndexError:
            pass
        try:
            n = args[5]
        except IndexError:
            pass
        else:
            jsq['limit'] = n
        jsq['tables'] = list(set(jsq['tables']))
        return SqlDict(jsq)


class SqlJoin(SqlFunc):
    """
    Функция создания соединения таблиц в запросе.
    """
    description = 'Функция создания соединения таблиц в запросе'
    args_description = [
        MandatoryArg('Первая таблица и поле', 0, [str]),
        MandatoryArg('Вторая таблица и поле', 1, [str]),
        NonMandatoryArg('Тип соединения', 2, [str]),
    ]

    @classmethod
    def get_name(self):
        return 'sql_join'

    def _operation(self, *args):
        try:
            type_ = args[2]
        except IndexError:
            type_ = 'inner'
        l, lf = args[0].split('.')
        r, rf = args[1].split('.')
        return SqlDict({
            'l': l,
            'r': r,
            'j': type_, 'on': {'l': lf, 'r': rf},
        })


class SqlWhere(SqlFunc):
    """
    Функция создания фильтров where в запросе.
    """
    description = 'Функция создания фильтров where в запросе'
    args_description = [
        MandatoryArg('Название таблицы и поля', 0, [str]),
        MandatoryArg('Тип фильтра', 1, [str]),
        MandatoryArg('Значение', 2, [str, int, float, list, SqlDict, datetime]),
    ]

    @classmethod
    def get_name(self):
        return 'sql_where'

    def _operation(self, *args):
        table_name, field_name = args[0].split('.')
        op, value = args[1], args[2]
        try:
            value = value.q_part
        except:
            pass
        return SqlWhereDict({
            'is_complex': False,
            'table': table_name,
            'field': field_name,
            'op': op,
            'value': value,
        })


class SqlOrderBy(SqlFunc):
    """
    Функция создания упорядочивания в запросе.
    """
    description = 'Функция создания упорядочивания в запросе'
    args_description = [
        MandatoryArg('Название таблицы и поля', 1, [str]),
        MandatoryArg('Направление', 0, [str]),
    ]

    @classmethod
    def get_name(self):
        return 'sql_orderby'

    def _operation(self, *args):
        table_name, field = args[1].split('.')
        ordering = args[0]
        return SqlDict({
            'table': table_name,
            'fields': [field],
            'type': ordering,
        })



class PlotlyMixin:
    """
    Доп. класс для компонентов-графиков plotly, при наследовании указывающий, что компонент
    является графиком на базе plotly, и реализующий методы
    отображения данных по графику в виде словаря и сохранения в файл.
    """
    @property
    def engine(self):
        return 'plotly'

    def to_dict(self):
        return {
            'data': self.figure.to_dict(),
            'title': self.title_text,
            'id': self.id,
        }

    def to_png(self, path):
        if not path.endswith('.png'):
            path += '.png'
        self.figure.write_image(path)

    def _no_groupby(self, dataframe, on, func):
        ser = dataframe.agg({on: func})
        df = ser.to_frame(name=func)
        df = df.reset_index()
        return df



"""
Для построения pdf-документа отчёта.
"""
from io import BytesIO

from reportlab.lib.pagesizes import A4, landscape
from reportlab.lib import colors
from reportlab.platypus import (
    Image,
    Table,
    PageBreak, 
    TableStyle,
    Paragraph,
    Spacer,
    SimpleDocTemplate,
)
from reportlab.lib.styles import ParagraphStyle
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.lib.units import inch


#  фикс для корректного отображения кириллицы
pdfmetrics.registerFont(TTFont('DejaVuSerif','DejaVuSerif.ttf', 'UTF-8'))


class PdfReport:
    """
    Класс конструктора отчёта в виде pdf.
    """
    def __init__(self, filename, **options):
        orientation = options.get('orientation', 'portrait')
        self._orientation = orientation
        if orientation == 'landscape':
            pagesize = landscape(A4)
        else:
            pagesize = A4
        self._doc = SimpleDocTemplate(filename=filename, pagesize=pagesize)

        #  "история" - список операций (классы из reportlab.platypus),
        #  которые выполняются при построении документа
        self._story = []

    def add_pagebreak(self):
        """
        Добавление разрыва страницы.
        """
        self._story.append(PageBreak())

    def add_table(self, data, **options):
        """
        Добавление таблицы.
        data - содержимое таблицы в виде списка списков -
        т.е. каждый список содержит значения ячеек соответствующей
        строки. Заголовок должен передаваться в этом списке первым элементом.
        """
        default_font = 'DejaVuSerif'
        default_font_size = 8

        title = options.get('title', '')
        title_font = options.get('title_font', 'DejaVuSerif')
        title_font_size = options.get('title_font_size', 14)
        font = options.get('font', default_font)
        font_size = options.get('font_size', default_font_size)
        table_style = TableStyle([
            ('LINEABOVE', (0, 0), (-1, -1), 1, colors.black),
            ('LINEBELOW', (0, 0), (-1, -1), 1, colors.black),
            ('FONT', (0, 0), (-1, -1), font),
            ('FONTSIZE', (0, 0), (-1, -1), font_size),
            ('GRID', (0, 0), (-1, -1), 1, colors.black),
        ])
        table = Table(data)
        table.setStyle(table_style)

        if title:
            style = ParagraphStyle(
                name='table_title',
                fontName=title_font,
                alignment=1,
                fontSize=title_font_size)
            self._story.append(Paragraph(text=title, style=style))
            self._story.append(Spacer(0, 10))

        self._story.append(table)
        self.add_pagebreak()
        return self

    def add_image(self, img):
        """
        Добавление изображения.
        img - изображение в виде последовательности байт.
        """
        image = Image(BytesIO(img))
        w, h = self._doc.width, self._doc.height
        if self._orientation == 'landscape':
            w, h = h, w
        image._restrictSize(w, h)
        self._story.append(image)
        self.add_pagebreak()
        return self

    def save(self):
        """
        Сохранение pdf-документа.
        """
        self._doc.build(self._story)

    def from_list(self, components_list):
        """
        Формирование документа из списка компонентов.
        Список должен содержать экземпляры классов компонентов
        отчёта. У экземпляров должен быть атрибут type, указывающий
        на тип компонента - таблица/график ('table', 'figure').
        У компонентов-таблиц должно быть свойство data, возвращающее
        список списков значений для таблицы. У компонентов-графиков
        должен быть метод to_bytes, возвращающий изображение графика
        в виде байт.
        """
        for component in components_list:
            type_ = component.type
            if type_ == 'table':
                _ = self.add_table(data=component.data, title=component.title_text)
            elif type_ == 'figure':
                _ = self.add_image(component.to_bytes())
        return self



from unittest.mock import patch, Mock

from noofa.core.sources.query import Table
from noofa.components.dataschema import (
    DataSchema,
    SchemaSource,
    SchemaQuery,
    SchemaDataframe
)
from noofa.tests.base import NoofaTest
from noofa.components.exceptions import SchemaComponentNotFound


class TestDataschema(NoofaTest):
    """
    Тестирование схемы данных.
    """
    @classmethod
    def setUpClass(cls):
        cls.sources = {
            'src1': {
                'id': 'src1',
                'type': 'postgres',
                'name': 'test_source',
            },
        }
        cls.queries = {
            'q1': {
                'id': 'q1',
                'name': 'test_query1',
                'build_from': 'json',
                'source': 'src1',
                'query_src': {
                    'base': 'table1',
                    'tables': ['table1'],
                    'filters': [
                        {
                            'is_complex': False,
                            'table': 'table1',
                            'field': 'col1',
                            'op': '>',
                            'value': 1,
                        }
                    ],
                },
            },
            'q2': {
                'id': 'q2',
                'name': 'test_query2',
                'build_from': 'expression',
                'source': 'src1',
                'query_src': 'sql_select("a")',
            },
            'q3': {
                'id': 'q3',
                'name': 'test_query3',
                'build_from': 'expression',
                'source': 'src2',
                'query_src': 'sql_select("a")',
            },
        }
        cls.dataframes = {
            'df1': {
                'id': 'df1',
                'name': 'test_df1',
                'base': {
                    'type': 'query',
                    'source': 'src1',
                    'value': 'q1',
                },
            },
            'df2': {
                'id': 'df2',
                'name': 'test_df2',
                'base': {
                    'type': 'source',
                    'source': '',
                    'value': 'src1',
                },
            },
            'df3': {
                'id': 'df3',
                'name': 'test_df3',
                'base': {
                    'type': 'expression',
                    'value': 'dataframe()',
                },
            },
        }

    def setUp(self):
        self.ds = DataSchema()

    def test_add_source(self):
        ds = self.ds
        self.assertRaises(SchemaComponentNotFound, ds.get_source, 'src1')

        n1 = len(ds._sources.keys())
        ds.add_source(**self.sources['src1'])
        n2 = len(ds._sources.keys())
        self.assertEqual(n1 + 1, n2)

        src = ds.get_source('src1')
        self.assertIsInstance(src, SchemaSource)

    def test_add_query(self):
        ds = self.ds

        self.assertRaises(SchemaComponentNotFound, ds.get_query, 'q1')
        self.assertRaises(SchemaComponentNotFound, ds.get_query, 'q2')

        ds.add_source(**self.sources['src1'])
        n1 = len(ds._queries.keys())
        ds.add_query(**self.queries['q1'])
        ds.add_query(**self.queries['q2'])
        n2 = len(ds._queries.keys())
        self.assertEqual(n1 + 2, n2)

        src = ds.get_query('q1')
        self.assertIsInstance(src, SchemaQuery)

        with self.assertRaises(KeyError):
            ds.add_query(**self.queries['q3'])

    def test_add_dataframe(self):
        ds = self.ds

        self.assertRaises(SchemaComponentNotFound, ds.get_dataframe, 'df1')
        self.assertRaises(SchemaComponentNotFound, ds.get_dataframe, 'df2')
        self.assertRaises(SchemaComponentNotFound, ds.get_dataframe, 'df3')

        ds.add_source(**self.sources['src1'])
        ds.add_query(**self.queries['q1'])
        n1 = len(ds._dataframes.keys())
        ds.add_dataframe(**self.dataframes['df1'])
        ds.add_dataframe(**self.dataframes['df2'])
        ds.add_dataframe(**self.dataframes['df3'])
        n2 = len(ds._dataframes.keys())
        self.assertEqual(n1 + 3, n2)

        df_names = list(ds._dataframes_by_name.keys())
        self.assertCountEqual(df_names, ['test_df1', 'test_df2', 'test_df3'])

        df = ds.get_dataframe('df1')
        self.assertIsInstance(df, SchemaDataframe)

    @patch('noofa.components.dataschema.SchemaSource', autospec=True)
    def test_schema_query(self, mock_schema_source):
        mock_source = Mock()
        mock_source.is_opened = True
        mock_source.get_table_multiple.return_value = {
            'table1': Table('table1', ['col1', 'col2'], enquote=True),
        }
        mock_conn = Mock()
        mock_conn.source_type = 'postgres'
        mock_source.connection = mock_conn
        mock_source.wildcard = '%s'
        mock_schema_source.return_value = mock_source

        ds = self.ds
        ds.add_source(**self.sources['src1'])
        ds.add_query(**self.queries['q1'])
        q = ds.get_query('q1')
        q.query = q.query_src
        query_obj = q._compile()
        expected_text = ''.join([
            'SELECT "table1"."col1", "table1"."col2" FROM table1 ',
            'WHERE "table1"."col1" > %s'
        ])
        self.assertEqual((expected_text, [1]), query_obj.str_and_params())

        mock_conn.source_type = 'mssql'
        mock_source.wildcard = '?'
        expected_text = expected_text.replace('%s', '?')
        query_obj = q._compile()
        self.assertEqual((expected_text, [1]), query_obj.str_and_params())

    @patch('noofa.components.dataschema.SchemaQuery', autospec=True)
    def test_schema_dataframe(self, mock_query):
        mock_data = Mock()
        mock_data.columns = ['a', 'b', 'c']
        mock_data.data = [(1, 2 ,3), (4, 5, 6)]
        mock_query.return_value.execute.return_value = mock_data

        ds = self.ds
        ds.add_source(**self.sources['src1'])
        ds.add_query(**self.queries['q1'])
        ds.add_dataframe(**self.dataframes['df1'])
        schema_df = ds.get_dataframe('df1')

        data = schema_df.get_data()
        self.assertTrue(hasattr(data, 'columns'))
        self.assertTrue(hasattr(data, 'data'))

        df = schema_df.build()
        self.assertEqual(df.shape, (2, 3))
        self.assertCountEqual(['a', 'b', 'c'], list(df.columns))



"""
Инструменты построения отчётов.
"""
import json
from io import BytesIO
from datetime import (
    date as dt_date,
    time as dt_time,
    datetime
)

from ..core.func.errors import InterpreterContextError
from ..core.func import Interpreter
from ..components.dataschema import DataSchema
from ..components.components import ComponentsSchema
from ..core.dataframes import panda_builder
from .exceptions import RecursiveDataframeBuildError


class ReportBuilder:
    """
    Формирователь отчётов.
    """
    def __init__(self, data_config=None, components_config=None, values=None, set_evaluator=True, *args, **kwargs):
        data_config = data_config or {}
        components_config = components_config or {}
        values = values or {}

        self._dataschema = DataSchema()  # схема данных
        self._components_schema = ComponentsSchema()  # схема компонентов
        self.interpreter = Interpreter()  # интерпретатор для вычисления формул
        self.interpreter._connections = self._dataschema._sources
        self._compiled_queries = {}  #  сформированные запросы
        self._built_dataframes = {}  #  сформированные датафреймы
        self._results = {}  #  результаты запросов (полученные данные)
        self._df_stack = []  #  стэк id строящихся датафреймов

        sources_conf = data_config.get('sources', {})
        queries_config = data_config.get('queries', {})
        dataframes_config = data_config.get('dataframes', {})

        # добавление значений
        self.interpreter.add_values(values.values(), self)

        """
        добавление источников в схему:
        источники добавляются из словаря (json) либо из строки подключения (conn_str)
        либо из выражения (expression);
        в случае выражения внешней функцией должна быть функция create_connection:
        например, 'create_connection(тип_соединения, строка_подключения)'
        """
        for s in sources_conf.values():
            build_from, value = s['from'], s['value']
            opts = {'id': s['id'], 'name': s['name'], 'type': s['type']}
            if build_from == 'json':
                opts.update(value)
            elif build_from == 'conn_str':
                opts['conn_str'] = value
            elif build_from == 'expression':
                conn = self.evaluate(value)
                opts['connection'] = conn
            self._dataschema.add_source(**opts)

        """
        добавление источников в схему:
        источники добавляются из словаря (json) либо из выражения (expression);
        выражение должно состоять из функции sql_select, например:
        'sql_select("table1", sql_join("table2", "table1", "field1", "field2"))'.
        """
        for q in queries_config.values():
            opts = {
                'id': q['id'], 'name': q['name'], 'source': q['source'],
                'build_from': q['from'], 'query_src': q['value']
            }
            self._dataschema.add_query(**opts)

        # добавление датафреймов в схему
        for df in dataframes_config.values():
            opts = {**df}
            self._dataschema.add_dataframe(**opts)

        # добавление компонентов в схему компонентов,
        # если параметр set_evaluator равен True, то для компонентов
        # в качестве параметра "вычислителя" задаётся сам ReportBuilder
        if set_evaluator == True:
            evaluator = self
        else:
            evaluator = None
        for c in components_config.values():
            type_ = c['type']
            if type_ == 'table':
                method = self._components_schema.add_table
            elif type_ == 'figure':
                method = self._components_schema.add_figure
            elif type_ == 'pivot_table':
                method = self._components_schema.add_pivot_table
            lo = c.get('layout', {})
            method(using_evaluator=evaluator, **c, **lo)

    def get_value(self, name):
        value = self.interpreter.get_value(name)
        return SingleValue(name, value)

    def evaluate(self, expr):
        """
        Вычисление значения по строке выражения.
        """
        try:
            return self.interpreter.evaluate(expr)
        except InterpreterContextError as e:
            df = self.get_or_build_dataframe(e.key)
            self.interpreter.add_to_global(e.key, df)
            return self.evaluate(expr)

    def apply(self, df_id, expr):
        """
        Применение выражения к строкам датафрейма.
        """
        df = self.get_or_build_dataframe(df_id)
        return self.interpreter.apply(df, expr)

    def apply_filters(self, filters):
        """
        Применить фильтры - т.е. задать для вычисляемых значений
        конфиги определённые значения.
        """
        for filter_key, filter_value in filters.items():
            self.interpreter.update_value(filter_key, filter_value)

    @property
    def dataframes(self):
        return self._dataschema._dataframes

    def build_query(self, query_id):
        """
        Сформировать объект запроса.
        """
        query = self._compiled_queries.get(query_id, None)
        if query is None:
            query = self.get_query(query_id)
            q = query._compile()
            self._compiled_queries[query.id] = q
            return q
        return query

    def get_data(self, query_id):
        """
        Получение данных по запросу.
        query_id - id запроса.
        Если данные имеются в self._result, то будут возвращены они,
        в противном случае выполняется запрос, после чего рез. вносится в self._results. 
        """
        if query_id in self._results:
            return self._results[query_id]
        query = self.get_query(query_id)
        qf = query.build_from
        if qf == 'json':
            query.query = query.query_src
        elif qf == 'expression':
            query.query = self.evaluate(query.query_src).q_part
        data = query.execute()
        self._compiled_queries[query.id] = query._compiled
        self._results[query.id] = data
        return data

    def get_or_build_dataframe(self, dataframe_id):
        """
        Получить сформированный либо сформировать датафрейм.
        """

        # проверка на рекурсивные вызовы при построении -
        # если в конфиге задано построение дф по своим же данным,
        # то выбрасывается ошибка
        if dataframe_id in self._df_stack:
            self._df_stack.clear()  # очистить стэк дф в случае ошибки
            raise RecursiveDataframeBuildError(dataframe_id)

        self._df_stack.append(dataframe_id)  # кладём id дф в стэк

        #  пробуем получить готовый дф, если его нет, то собираем с нуля
        df = self._built_dataframes.get(dataframe_id, None)
        if df is None:
            df = self.build_dataframe(dataframe_id)

        #  добавляем в словарь готовых дф для возможного последующего использования
        self._built_dataframes[dataframe_id] = df

        # удаляем id дф из стэка
        self._df_stack.pop()

        return df

    def build_base(self, dataframe_id):
        """
        Построение основы датафрейма.

        dataframe_id - id датафрейма.
        """
        df = self.get_dataframe(dataframe_id)
        build_type = df.build_type
        if build_type == 'query':
            res = self.get_data(df._query.id)
            dataframe = panda_builder.new(res.data, res.columns)
        elif build_type == 'expression':
            dataframe = self.evaluate(df.build_from)
        elif build_type == 'source':
            dataframe = df.get_data()
        return dataframe

    def build_dataframe(self, dataframe_id):
        """
        Сформировать датафрейм.

        dataframe_id - id датафрейма.
        """
        df = self.get_dataframe(dataframe_id)
        #  создание базового датафрейма - либо из запроса, либо из выражения
        dataframe = self.build_base(dataframe_id)

        # преобразование типов
        for dt in df.dtypes:
            try:
                col, dtype = dt['col'], dt['dtype']
                dataframe = panda_builder.astype(dataframe, col, dtype)
            except:
                pass

        #  приклеивание других датафреймов
        for u in df.unions:
            from_, value = u['from'], u['value']
            if from_ == 'expression':
                df2 = self.evaluate(value)
            else:
                df2 = self.get_or_build_dataframe(value)
            dataframe = panda_builder.union([dataframe, df2])

        #  добавление соединений с другими датафреймами
        for j in df.joins:
            from_, value = j['from'], j['value']
            on, join_type = j['on'], j['type']
            if from_ == 'expression':
                df2 = self.evaluate(value)
            else:
                df2 = self.get_or_build_dataframe(value)
            dataframe = panda_builder.join(dataframe, df2, on, join_type)

        #  добавление новых столбцов либо изменение существующих
        extra_cols_n, extra_cols = len(df.cols), {}
        for col in df.cols:
            from_, col_name, expr = col['from'], col['name'], col['value']
            if from_ == 'expression':
                value = self.evaluate(expr)
            elif from_ == 'apply':
                value = self.interpreter.apply(dataframe, expr)
            if extra_cols_n == 1:
                dataframe = panda_builder.add_column(dataframe, col_name, value)
            else:
                extra_cols[col_name] = value
        if extra_cols:
            dataframe = panda_builder.add_columns(dataframe, extra_cols)

        #  применение фильтров
        filters = []
        for f in df.filters:
            from_, filter_ = f['from'], f['value']
            if from_ == 'expression':
                filter_ = self.evaluate(filter_).df_filter
            filters.append(filter_)
        if filters:
            dataframe = panda_builder.filter(dataframe, filters)

        #  упорядочивание
        orderings = {'cols': [], 'asc': []}
        for ordering in df.ordering:
            cols, asc = ordering['cols'], ordering['asc']
            orderings['cols'].append(cols)
            orderings['asc'].append(asc)
        if orderings['cols']:
            cols, asc = orderings['cols'], orderings['asc']
            dataframe = panda_builder.order_by(dataframe, cols, asc=asc)

        # обработка пустых значений
        for fn in df.fillna:
            col, action, value = fn['col'], fn['action'], fn['value']
            if action == 'drop':
                dataframe = panda_builder.drop_na(dataframe, col)
            elif action == 'fill':
                fill_value = self.evaluate(value)
                dataframe = panda_builder.fill_na(dataframe, col, fill_value)

        return dataframe

    def build_table(self, table, **kwargs):
        """
        Формирование компонента-таблицы.
        table - id компонента в наборе таблиц схемы компонентов
        либо экземпляр компонента-таблицы.
        """
        if isinstance(table, str):
            table = self._components_schema.get_table(table)
        #  если у компонента нет "объекта-вычислителя",
        #  то им становится сам ReportBuilder
        if table.evaluator is None:
            table.evaluator = self
        table.build(**kwargs)
        return table

    def build_figure(self, figure):
        """
        Формирование компонента-графика.
        figure - id компонента в наборе графиков схемы компонентов
        либо экземпляр компонента-графика.
        """
        if isinstance(figure, str):
            figure = self._components_schema.get_figure(figure)
        if figure.evaluator is None:
            figure.evaluator = self
        _ = figure.build()
        return figure

    def build_component(self, component):
        if isinstance(component, str):
            component = self.get_component(component)
        method = getattr(self, f'build_{component.type}')
        return method(component)

    def bufferize_xlsx(self, cmp_type, cmp_id, **kwargs):
        """
        Получение содержимого результата запроса, датафрейма либо таблицы
        в виде буфера байт.
        """
        buffer = BytesIO()
        is_pt = False
        if cmp_type == 'dataframe':
            df = self.get_or_build_dataframe(cmp_id)
        elif cmp_type == 'query':
            data = self.get_data(cmp_id)
            df = panda_builder.new(data.data, data.columns)
        elif cmp_type == 'table':
            table = self.build_table(cmp_id, **kwargs)
            is_pt = hasattr(table, 'is_pivot_table')
            df = table.pivot_df if is_pt else table.df
        df.to_excel(buffer, index=is_pt)
        return buffer

    def df_to_dict(self, dataframe_id):
        """
        Датафрейм -> словарь.
        """
        df = self.get_or_build_dataframe(dataframe_id)
        return df.to_dict(orient='records')

    def get_source(self, source_id):
        return self._dataschema.get_source(source_id)

    def get_query(self, query_id):
        return self._dataschema.get_query(query_id)

    def get_dataframe(self, dataframe_id):
        return self._dataschema.get_dataframe(dataframe_id)

    def get_component(self, component_id):
        return self._components_schema.get_component(component_id)


class SingleValue:
    def __init__(self, name, value):
        self.name = name
        self.value = value

    @property
    def is_simple(self):
        stypes = [
            str, int, float, bool, list,
            datetime, dt_date, dt_time,
        ]
        return type(self.value) in stypes



import json
from datetime import datetime

from ..core.sources.conn import DataQueryResult
from ..core.dataframes import panda_builder
from .builders import ReportBuilder
from .enc import NoofaJsonEncoder


class CacheAgent:
    """
    Средство для подготовки к кэшированию и "декэширования"
    датафреймов и результатов запросов.
    """
    _meta_label = '_meta'

    def __init__(self, data, meta=None):
        self._data = data

        """
        доп. инф. о данных - используется
        для учёта полей даты и времени, которые
        были преобразованы в строки для кэширования и
        потребуют обратного преобразования после получения из кэша
        """
        self._meta = meta

        _DF = panda_builder.pd.DataFrame
        if not isinstance(self._data, _DF):
            self._data = _DF(self._data)

    @classmethod
    def from_cached(cls, cached, **kwargs):
        """
        Создание экзмепляра с использованием полученных из redis данных.
        """
        data = json.loads(cached)
        return cls(
            data['data'],
            data[cls._meta_label],
            **kwargs,
        )

    def load(self, df=True):
        """
        Преобразование полученных из redis данных в датафрейм
        либо словаря с результатами запроса.
        При df=True возвращается датафрейм, при False - словарь с рез. запроса.
        """
        dt_columns = self._meta.get('datetime_columns', [])
        data = self._data.copy(deep=True)
        if dt_columns and not data.empty:
            data[dt_columns] = data[dt_columns].astype('datetime64')
        if df == True:
            return data
        return {
            'data': data.to_dict(orient='records'),
            'columns': list(data.columns.values)
        }

    def dump(self):
        """
        Подготовка данных и преобразование их в строку перед кэшированием
        в redis. При подготовке данных значения столбцов с датой и временем
        преобразуются в строки.
        """
        meta = {'datetime_columns': []}
        df = self._data.copy(deep=True)
        dt_columns = df.select_dtypes(include=['datetime', 'datetimetz'])
        dt_columns = list(dt_columns.columns.values)
        if dt_columns:
            df[dt_columns] = df[dt_columns].astype('str')
            meta['datetime_columns'] = dt_columns
        data = df.to_dict(orient='records')
        return json.dumps({
            'data': data,
            self.__class__._meta_label: meta,
        }, cls=NoofaJsonEncoder)


class CachingReportBuilder(ReportBuilder):
    """
    Формирователь отчётов с функционалом кэширования в redis.
    Это дочерний класс от ReportBuilder с измененными методами get_or_build_dataframe и
    get_data и несколькими дополнительными методами, связанными с кэшированием.
    При создании датафрейма кэшируется его содержимое в виде словаря. При построении
    датафрейма сначала делается попытка получить кэшированное содержимое для датафрейма,
    в случае отсутствия такового - выполняется построение с нуля.
    """
    def __init__(
        self,
        redis_connection=None,
        redis_prefix='',
        redis_expire=300,
        storing_only=False,
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)

        # соединение с redis
        self._conn = redis_connection

        # префикс для ключей в redis
        self._prefix = redis_prefix if redis_prefix else '__'

        # время хранения значения в кэше
        self._expire = redis_expire

        # режим:
        # 0 - только сохранение в кэш (без обращения к кэшу для получения готовых данных по запросам/датафреймам),
        # 1 - обращение за результатами к кэшу + кэширование новых результатов
        self._mode = 0 if storing_only == True else 1

    def storing_only(self):
        """
        Включить режим только сохранения.
        """
        self._mode = 0

    @property
    def is_storing_only(self):
        return self._mode == 0

    def get_or_build_dataframe(self, dataframe_id):
        from_cache = False

        #  сначала пробуем получить уже готовый дф
        df = self._built_dataframes.get(dataframe_id, None)
        #  если его нет, то пробуем построить его по данным из кэша
        if df is None:
            cached_df = None if self.is_storing_only else self.get_cached_df(dataframe_id)
            if cached_df is not None:
                from_cache = True
                df = cached_df
            #  если в кэше данных нет - строим дф заново
            else:
                df = self.build_dataframe(dataframe_id)

        #  добавляем в словарь готовых дф для возможного последующего использования,
        #  чтобы не обращаться дополнительно к кэшу
        self._built_dataframes[dataframe_id] = df

        #  кэшируем дф для возможного последующего использования его данных другими
        #  экземплярами CachingReportBuilder
        if not from_cache:
            self.cache_df(dataframe_id, df)

        return df

    def get_data(self, query_id):
        from_cache = False

        #  сначала проверяем наличие рез. запроса в сохраненных результатах -
        #  в self._results
        if query_id in self._results:
            data = self._results[query_id]
        else:
            #  если их там нет - пробуем получить из кэша
            data = None if self.is_storing_only else self.get_cached_query_result(query_id)

        #  если в кэше нет результата - выполняем запрос заново
        if data is None:
            data = super().get_data(query_id)
        else:
            from_cache = True
            data = DataQueryResult(**data)

        #  кэшируем для возможного послед. использования
        if not from_cache:
            self.cache_query_result(query_id, data.data)

        return data

    def cache_df(self, dataframe_id, df, ex=None):
        """
        Кэшировать датафрейм.
        dataframe_id - id датафрейма,
        df - экз. pandas.DataFrame либо словарь,
        ex - время хранения в кэше в секундах.
        """
        ca = CacheAgent(df)
        data = ca.dump()
        self._set(
            dataframe_id,
            data,
            'dataframe',
            self._get_expire(ex)
        )

    def get_cached_df(self, dataframe_id):
        """
        Получение кэшированных данных датафрейма.
        dataframe_id - id датафрейма.
        """
        cached_df = self._get(dataframe_id, 'dataframe')
        if cached_df:
            ca = CacheAgent.from_cached(cached_df)
            df = ca.load(df=True)

            # чтобы не было ошибок при получении пустого дф из кэша
            cols = list(df.columns)
            df = df if len(cols) else None

            return df

    def cache_query_result(self, query_id, query_result, ex=None):
        """
        Кэшировать результат запроса.
        query_id - id запроса,
        query_result - результат запроса в виде списка словарей,
        ex - время хранения в кэше в секундах.
        """
        ca = CacheAgent(query_result)
        data = ca.dump()
        self._set(
            query_id,
            data,
            'query',
            self._get_expire(ex)
        )

    def get_cached_query_result(self, query_id):
        """
        Получение кэшированных результатов запроса.
        query_id - id запроса.
        """
        cached_qr = self._get(query_id, 'query')
        if cached_qr:
            ca = CacheAgent.from_cached(cached_qr)
            qr = ca.load(df=False)
            return qr

    def _set(self, key, value, value_type, ex=None, *args, **kwargs):
        """
        Установка значения в кэш.
        """
        self._conn.set(
            self._key(key, value_type),
            value,
            ex=self._get_expire(ex)
        )

    def _get(self, key, value_type):
        """
        Получение значения из кэша.
        """
        return self._conn.get(
            self._key(key, value_type)
        )

    def clear(self, key, value_type):
        self._conn.delete(self._key(key, value_type))

    def _key(self, key, value_type):
        return f'{self._prefix}:{value_type}:{key}'

    def _get_expire(self, expire):
        return expire if expire is not None else self._expire



_TEST_QUERIES = {
    'postgres': 'SELECT 1',
    'mysql': 'SELECT 1',
    'sqlite': 'SELECT 1',
    'mssql': 'SELECT 1',
}


_TABLES_QUERIES = {
    'postgres': "SELECT table_name FROM information_schema.tables " + \
            "WHERE table_schema NOT IN ('pg_catalog', 'information_schema') ORDER BY table_name;",

    'mysql': "SELECT table_name FROM information_schema.tables " + \
            "WHERE table_schema = %s ORDER BY table_name",

    'mssql': 'SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES;',

    'sqlite': "SELECT name FROM sqlite_master WHERE type='table';",
}


_COLUMNS_QUERIES = {
    'postgres': "SELECT table_name, column_name, data_type FROM information_schema.columns " + \
            "WHERE table_schema NOT IN ('pg_catalog', 'information_schema') ORDER BY table_name, column_name;",

    'mysql': "SELECT table_name, column_name, column_type from information_schema.columns " + \
            "WHERE table_schema = %s ORDER BY table_name, column_name;",

    'mssql': "SELECT table_name, column_name, data_type FROM information_schema.columns " + \
            "ORDER BY table_name, column_name;",
}


_FIELDS_QUERIES = {
    'postgres': "SELECT column_name FROM information_schema.columns " + \
            "WHERE table_name = %s AND table_schema NOT IN ('pg_catalog', 'information_schema') " + \
            "ORDER BY column_name;",

    'mysql': "SELECT column_name FROM information_schema.columns " + \
            "WHERE table_name = %s AND table_schema = %s ORDER BY column_name;",

    'mssql': "SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS " + \
            "WHERE TABLE_NAME = {?} ORDER BY COLUMN_NAME;",

    'sqlite': "SELECT * from %s;",
}


_PG_MSSQL_CONSTR_Q = "SELECT tc.table_name as referencing_table, " + \
        "tc.constraint_type as constr_type, " + \
        "ksu.column_name as col_name, " + \
        "ccu.table_name as referenced_table, " + \
        "ccu.column_name as referenced_column " + \
        "FROM information_schema.table_constraints as tc " + \
        "INNER JOIN information_schema.key_column_usage as ksu " + \
        "ON tc.constraint_name = ksu.constraint_name " + \
        "AND tc.table_schema = ksu.table_schema " + \
        "INNER JOIN information_schema.constraint_column_usage as ccu " + \
        "ON tc.constraint_name = ccu.constraint_name " + \
        "AND tc.table_schema = ccu.table_schema " + \
        "WHERE tc.constraint_type in ('PRIMARY KEY', 'FOREIGN KEY') " + \
        "ORDER BY constr_type, referencing_table"


_MYSQL_CONSTR_Q = "SELECT tc.table_name as referencing_table, " + \
        "tc.constraint_type as constr_type, " + \
        "kcu.column_name as col_name, " + \
        "kcu.referenced_table_name as referenced_table, " + \
        "kcu.referenced_column_name as referenced_column " + \
        "FROM information_schema.table_constraints as tc " + \
        "INNER JOIN information_schema.key_column_usage as kcu " + \
        "ON tc.constraint_name = kcu.constraint_name " + \
        "AND tc.table_schema = kcu.table_schema " + \
        "AND tc.table_name = kcu.table_name " + \
        "WHERE tc.constraint_type in ('PRIMARY KEY', 'FOREIGN KEY') " + \
        "AND tc.table_schema = %s"


_CONSTRAINTS_QUERIES = {
    'postgres': _PG_MSSQL_CONSTR_Q,
    'mssql': _PG_MSSQL_CONSTR_Q,
    'mysql': _MYSQL_CONSTR_Q,
}



from noofa.tests.base import NoofaSuite, NoofaRunner
from noofa.tests.df.panda_filters import TestPandaFilters
from noofa.tests.df.panda_builder import TestPandaBuilder


def df_suite():
    suite = NoofaSuite()
    suite.add(TestPandaFilters)
    suite.add(TestPandaBuilder)
    return suite


def run_tests():
    suite = df_suite()
    runner = NoofaRunner()
    runner.run(suite)



import unittest

from noofa.core.func.interpreter import Interpreter
from noofa.core.func.errors import NotEnoughArguments as NEA
from noofa.core.func.errors import ArgumentTypeError as ATE


class NoofaTest(unittest.TestCase):
    @property
    def is_noofa(self):
        return True

    @classmethod
    def suite(cls):
        return unittest.makeSuite(cls)


class NoofaSuite(unittest.TestSuite):
    def add(self, test_component):
        is_noofa_test = False
        try:
            is_noofa_test = test_component.is_noofa
        except AttributeError:
            pass
        if is_noofa_test:
            self.addTest(test_component.suite())
        else:
            _ = test_component
            if isinstance(test_component, unittest.TestCase):
                _ = unittest.makeSuite(test_component)
            self.addTest(_)


class NoofaRunner(unittest.TextTestRunner):
    pass


class NoofaInterpreterTest(NoofaTest):
    @classmethod
    def set_up_interpreter(cls):
        cls.interpreter = Interpreter()

    def evaluate(self, expression):
        return self.interpreter.evaluate(expression)


class NoofaFunctionTest(NoofaTest):
    @classmethod
    def add_cases_from_list(cls, cases=[]):
        cls.cases = FunctionCase.from_list(cases)

    def test_functionality(self):
        for fc in self.cases:
            cases = fc.get_subcases()
            for case in cases:
                try:
                    f, expected, assertion = case
                except ValueError:
                    f, expected = case
                    assertion = 'equal'

                msg = fc.func.get_name()
                with self.subTest(msg):
                    if assertion == 'equal':
                        self.assertEqual(f(), expected)
                    elif assertion == 'isinstance':
                        self.assertIsInstance(f(), expected)

    def test_nea(self):
        for fc in self.cases:
            case = fc.nea_case
            if case is None:
                continue
            with self.subTest():
                self.assertRaises(NEA, case)

    def test_ate(self):
        for fc in self.cases:
            for case in fc.ate_cases:
                with self.subTest(case):
                    self.assertRaises(ATE, case)


class FunctionCase:
    def __init__(self, func, subcases=(), ate=(), nea_args=()):
        self.func = func
        self.subcases = subcases
        self.nea_args = nea_args
        self.ate = ate

    def get_subcases(self):
        cases = []
        for sc in self.subcases:
            args, expected = sc[0], sc[1]
            try:
                assertion = sc[2]
            except IndexError:
                assertion = 'equal'
            if not isinstance(args, tuple):
                args = (args, )
            cases.append((self.func(*args), expected, assertion))
        return cases

    @property
    def nea_case(self):
        if self.nea_args is not None:
            return self.func(*self.nea_args)

    @property
    def ate_cases(self):
        ate_cases = []
        for args in self.ate:
            if not isinstance(args, tuple):
                args = (args, )
            ate_cases.append(self.func(*args))
        return ate_cases

    @classmethod
    def from_list(cls, cases):
        return [cls(*case) for case in cases]



from noofa.tests.base import NoofaSuite, NoofaRunner
from noofa.tests.sources.conn import TestParseConnString
from noofa.tests.sources.query_build import (
    TestQueryComponents,
    TestQueryPreparation,
    TestQueryBuild,
    TestQueryFilters,
)


def sources_suite():
    suite = NoofaSuite()
    suite.add(TestParseConnString)
    suite.add(TestQueryComponents)
    suite.add(TestQueryPreparation)
    suite.add(TestQueryBuild)
    suite.add(TestQueryFilters)
    return suite


def run_tests():
    suite = sources_suite()
    runner = NoofaRunner()
    runner.run(suite)



from noofa.tests.base import NoofaSuite, NoofaRunner
from noofa.tests.sources import sources_suite
from noofa.tests.df import df_suite
from noofa.tests.func import func_suite
from noofa.tests.components import components_suite
from noofa.tests.builders import builders_suite


def run_tests():
    suite = NoofaSuite()
    suite.add(sources_suite())
    suite.add(df_suite())
    suite.add(func_suite())
    suite.add(components_suite())
    suite.add(builders_suite())
    runner = NoofaRunner()
    runner.run(suite)


if __name__ == '__main__':
    run_tests()




from datetime import date, time
from json import JSONEncoder


class NoofaJsonEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (date, time)):
            return obj.isoformat()
        return JSONEncoder.default(self, obj)



from noofa.tests.base import NoofaFunctionTest
from noofa.core.func import sql


class TestSqlFunctions(NoofaFunctionTest):
    """
    Тестирование функций построения sql-запросов.
    """
    @classmethod
    def setUpClass(cls):
        cls.add_cases_from_list(_CASES)

    def test_select(self):
        f1 = sql.SqlSelect('table')
        sql_dict = f1()
        expected_keys = [
            'is_subquery', 'tables', 'base',
            'joins', 'filters', 'values', 'order_by'
        ]
        self.assertCountEqual(sql_dict.keys(), expected_keys)

        expected_keys.append('limit')
        f2 = sql.SqlSelect('table', [], [], [], [], 10)
        sql_dict = f2()
        self.assertCountEqual(sql_dict.keys(), expected_keys)

    def test_join(self):
        f1 = sql.SqlJoin('table1.col1', 'table2.col2')
        sql_dict = f1()
        expected_keys = ['l', 'r', 'j', 'on']
        self.assertCountEqual(sql_dict.keys(), expected_keys)
        self.assertEqual(sql_dict['j'], 'inner')

        f2 = sql.SqlJoin('table1.col1', 'table2.col2', 'left')
        sql_dict = f2()
        self.assertEqual(sql_dict['j'], 'left')

    def test_where(self):
        f = sql.SqlWhere('table.col', '>', 0)
        sql_dict = f()
        expected_keys = ['is_complex', 'table', 'field', 'op', 'value']
        self.assertCountEqual(sql_dict.keys(), expected_keys)

    def test_orderby(self):
        f = sql.SqlOrderBy('asc', 'table.col')
        sql_dict = f()
        expected_keys = ['table', 'fields', 'type']
        self.assertCountEqual(sql_dict.keys(), expected_keys)

    def test_complex_where(self):
        expected_keys = ['is_complex', 'op', 'filters']

        f1 = sql.SqlWhere('table.col', '>', 0)()
        f2 = sql.SqlWhere('table.col', '<', 1)()
        complex_where1 = f1 & f2
        self.assertCountEqual(complex_where1.keys(), expected_keys)
        self.assertCountEqual([f1._q, f2._q], complex_where1['filters'])
        complex_where2 = f1 | f2
        self.assertCountEqual(complex_where2.keys(), expected_keys)
        self.assertCountEqual([f1._q, f2._q], complex_where2['filters'])

"""
Содержимое кортежей в списке _CASES:
(
    Функция,
    сабкейсы (кортеж из аргументов и ожидаемого результата),
    аргументы для проверки ошибки по типу аргумента,
    аргументы для проверки ошибки по кол-ву аргументов
)
"""
_CASES = (
    (sql.SqlSelect, ((('table', ), sql.SqlDict, 'isinstance'), ), ((1, ), ), ()),
    (sql.SqlJoin,
        ((('table.col', 'table1.col', 'inner'), sql.SqlDict, 'isinstance'), ),
        ((1, 2), ),
        (('table', ),)
    ),
    (sql.SqlWhere,
        ((('table.col', '>', 0), sql.SqlWhereDict, 'isinstance'), ),
        ((1, '', ''), ),
        ('table.col',)
     ),
    (sql.SqlOrderBy,
        ((('asc', 'table.col'), sql.SqlDict, 'isinstance'), ),
        (('asc', 0), ),
        ('asc', )
     ),
)



from noofa.tests.base import NoofaSuite, NoofaRunner
from noofa.tests.func.interpreter import (
    TestInterpreter,
    TestOperators,
    TestOperatorsClasses,
)
from noofa.tests.func.context import TestContext
from noofa.tests.func.math import TestMathFunctions
from noofa.tests.func.datastruct import TestDatastructFunctions
from noofa.tests.func.date import TestDateFunctions
from noofa.tests.func.typeconv import TestTypeConvFunctions
from noofa.tests.func.logic import TestLogicFunctions
from noofa.tests.func.strings import TestStringsFunctions
from noofa.tests.func.df import TestDfFunctions
from noofa.tests.func.sql import TestSqlFunctions
from noofa.tests.func.statistics import TestStatisticsFunctions
from noofa.tests.func.expr import TestFunctionsExpressions


def func_suite():
    suite = NoofaSuite()
    suite.add(TestOperators)
    suite.add(TestOperatorsClasses)
    suite.add(TestInterpreter)
    suite.add(TestContext)
    suite.add(TestMathFunctions)
    suite.add(TestDatastructFunctions)
    suite.add(TestDateFunctions)
    suite.add(TestTypeConvFunctions)
    suite.add(TestLogicFunctions)
    suite.add(TestStringsFunctions)
    suite.add(TestSqlFunctions)
    suite.add(TestStatisticsFunctions)
    suite.add(TestDfFunctions)
    suite.add(TestFunctionsExpressions)
    return suite


def run_tests():
    suite = func_suite()
    runner = NoofaRunner()
    runner.run(suite)



from noofa.tests.base import NoofaTest
from noofa.core.sources.conn import _parse_conn_str


class TestParseConnString(NoofaTest):
    """
    Тестирование парсинга строки подключения.
    """
    def setUp(self):
        self._wrong_param = 'wrong_param'
        self.conn_str = 'host=host;port=1;database=db;'
        self.conn_str += f'user=user;password=pwd;ad=true'
        self.conn_str_with_wrong_param = self.conn_str + f';{self._wrong_param}=true'
        self.bad_conn_str = 'param1=1;param2'

    def test_parse(self):
        parsed = _parse_conn_str(self.conn_str)
        for param in ['host', 'port', 'user', 'password', 'database', 'ad']:
            self.assertIn(param, parsed)

    def test_parse_with_wrong_param(self):
        parsed = _parse_conn_str(self.conn_str_with_wrong_param)
        self.assertNotIn(self._wrong_param, parsed)

    def test_bad_conn_str(self):
        self.assertRaises(Exception, _parse_conn_str, self.bad_conn_str)



"""
Операторы.
"""
from datetime import datetime

from .base import Operator, MandatoryArg
from ..dataframes import panda_builder


_OPERATORS_PRIORITY = {
    '*': 1,
    '/': 1,
    '+': 2,
    '-': 2,
    '>': 3,
    '<': 3,
    '>=': 3,
    '<=': 3,
    '!=': 4,
    '==': 4,
    '&': 5,
    '|': 6,
}


_NONE_TYPE = type(None)


class Add(Operator):
    """
    Сумма двух чисел.
    """
    sign = '+'
    description = 'Сумма двух чисел или конкатенация строк'
    args_description = [
        MandatoryArg('Число1', 0, [int, float, str, _NONE_TYPE, panda_builder.pd.DataFrame]),
        MandatoryArg('Число2', 1, [int, float, str, panda_builder.pd.DataFrame]),
    ]

    def _operation(self, *args):
        left, right = args
        if isinstance(left, panda_builder.pd.DataFrame) and isinstance(right, panda_builder.pd.DataFrame):
            return panda_builder.union([left, right])
        if left is None and type(right) is not str:
            return right
        return left + right


class Subtract(Operator):
    """
    Вычитание.
    """
    sign = '-'
    description = 'Вычитание'
    args_description = [
        MandatoryArg('Число1', 0, [int, float, str, _NONE_TYPE]),
        MandatoryArg('Число2', 1, [int, float, str]),
    ]

    def _operation(self, *args):
        left, right = args
        if left is None and type(right) is not str:
            left = 0
        return left - right


class Divide(Operator):
    """
    Деление.
    """
    sign = '/'
    description = 'Деление'
    args_description = [
        MandatoryArg('Число1', 0, [int, float]),
        MandatoryArg('Число2', 1, [int, float]),
    ]

    def _operation(self, *args):
        return args[0] / args[1]


class Multiply(Operator):
    """
    Умножение.
    """
    sign = '*'
    description = 'Умножение'
    args_description = [
        MandatoryArg('Число1', 0, [int, float]),
        MandatoryArg('Число2', 1, [int, float]),
    ]

    def _operation(self, *args):
        return args[0] * args[1]


class Assign(Operator):
    """
    Присваивание значения.
    """
    sign = '='
    description = 'Присваивание значения'
    args_description = [
        MandatoryArg('Имя переменной', 0),
        MandatoryArg('Значение', 1),
    ]

    def _operation(self, *args):
        return {args[0]: args[1]}


class IsGt(Operator):
    """
    Сравнение >.
    """
    sign = '>'
    description = 'Сравнение > двух значений'
    args_description = [
        MandatoryArg('Значение1', 0, [int, float, bool, datetime]),
        MandatoryArg('Значение2', 1, [int, float, bool, datetime]),
    ]

    def _operation(self, *args):
        return args[0] > args[1]


class IsGte(Operator):
    """
    Сравнение >=.
    """
    sign = '>='
    description = 'Сравнение >= двух значений'
    args_description = [
        MandatoryArg('Значение1', 0, [int, float, bool, datetime]),
        MandatoryArg('Значение2', 1, [int, float, bool, datetime]),
    ]

    def _operation(self, *args):
        return args[0] >= args[1]


class IsLt(Operator):
    """
    Сравнение <.
    """
    sign = '<'
    description = 'Сравнение < двух значений'
    args_description = [
        MandatoryArg('Значение1', 0, [int, float]),
        MandatoryArg('Значение2', 1, [int, float]),
    ]

    def _operation(self, *args):
        return args[0] < args[1]


class IsLte(Operator):
    """
    Сравнение <=.
    """
    sign = '<='
    description = 'Сравнение <= двух значений'
    args_description = [
        MandatoryArg('Значение1', 0, [int, float]),
        MandatoryArg('Значение2', 1, [int, float]),
    ]

    def _operation(self, *args):
        return args[0] <= args[1]


class IsEq(Operator):
    """
    Сравнение ==.
    """
    sign = '=='
    description = 'Сравнение == двух значений'
    args_description = [
        MandatoryArg('Значение1', 0, [str, int, float, bool, datetime, list]),
        MandatoryArg('Значение2', 1, [str, int, float, bool, datetime, list]),
    ]

    def _operation(self, *args):
        return args[0] == args[1]


class IsNeq(Operator):
    """
    Сравнение !=.
    """
    sign = '!='
    description = 'Сравнение != двух значений'
    args_description = [
        MandatoryArg('Значение1', 0, [str, int, float, bool, datetime, list]),
        MandatoryArg('Значение2', 1, [str, int, float, bool, datetime, list]),
    ]

    def _operation(self, *args):
        return args[0] != args[1]


class Or(Operator):
    """
    Логическое ИЛИ
    """
    sign = '|'
    description = 'Логическое ИЛИ'
    args_description = [
        MandatoryArg('Значение1', 0),
        MandatoryArg('Значение2', 1),
    ]

    def _operation(self, *args):
        left, right = args
        return left | right


class And(Operator):
    """
    Логическое И
    """
    sign = '&'
    description = 'Логическое И'
    args_description = [
        MandatoryArg('Значение1', 0),
        MandatoryArg('Значение2', 1),
    ]

    def _operation(self, *args):
        left, right = args
        return left & right



"""
Функции работы с датой и временем.
"""

from datetime import datetime, date, time

from .base import DateFunc, NonMandatoryArg, MandatoryArg


_DEFAULT_FMT = '%Y-%m-%d %H:%M:%S'
_DEFAULT_TIME_FMT = '%H:%M:%S'
_DEFAULT_DATE_FMT = '%Y-%m-%d'

_DEFAULT_FMTS = {
    time: _DEFAULT_TIME_FMT,
    date: _DEFAULT_DATE_FMT,
    datetime: _DEFAULT_FMT,
}


def _seconds_to_datetimestr(seconds, fmt):
    return datetime.fromtimestamp(seconds).strftime(fmt)


class SecToDate(DateFunc):
    """
    Перевод секунд в дату и время.
    """
    description = 'Перевод секунд в дату и время'
    args_description = [
        MandatoryArg('Секунды', 0, [int, float]),
    ]

    def _operation(self, *args):
        seconds = args[0]
        return datetime.fromtimestamp(seconds)


class MsecToDate(DateFunc):
    """
    Перевод миллисекунд в дату и время.
    """
    description = 'Перевод миллисекунд в дату и время'
    args_description = [
        MandatoryArg('Миллисекунды', 0, [int, float]),
    ]

    def _operation(self, *args):
        msec = args[0]
        return datetime.fromtimestamp(msec/1000)


class SecToDateStr(DateFunc):
    """
    Перевод секунд в строку даты и времени.
    """
    description = 'Перевод секунд в строку даты и времени'
    args_description = [
        MandatoryArg('Секунды', 0, [int, float]),
        NonMandatoryArg('Формат', 1, [str]),
    ]

    def _operation(self, *args):
        seconds = args[0]
        try:
            fmt = args[1]
        except IndexError:
            fmt = _DEFAULT_FMT
        return _seconds_to_datetimestr(seconds, fmt)


class MsecToDateStr(DateFunc):
    """
    Перевод миллисекунд в строку даты и времени.
    """
    description = 'Перевод миллисекунд в строку даты и времени'
    args_description = [
        MandatoryArg('Миллисекунды', 0, [int, float]),
        NonMandatoryArg('Формат', 1, [str]),
    ]

    def _operation(self, *args):
        seconds = args[0]
        try:
            fmt = args[1]
        except IndexError:
            fmt = _DEFAULT_FMT
        return _seconds_to_datetimestr(seconds/1000, fmt)


class Now(DateFunc):
    """
    Получение текущей даты и времени.
    """
    description = 'Получение текущей даты и времени'
    args_description = []

    def _operation(self, *args):
        return datetime.now()


class Date(DateFunc):
    """
    Получение текущей даты.
    """
    description = 'Получение текущей даты'
    args_description = []

    def _operation(self, *args):
        return datetime.now().date()


class Time(DateFunc):
    """
    Получение текущего времени.
    """
    description = 'Получение текущего времени'
    args_description = []

    def _operation(self, *args):
        return datetime.now().time()


class DateToSec(DateFunc):
    """
    Перевод даты в секунды.
    """
    description = 'Перевод даты в секунды'
    args_description = [
        MandatoryArg('Дата', 0, [datetime]),
    ]

    def _operation(self, *args):
        return int(args[0].timestamp())


class DateToStr(DateFunc):
    """
    Перевод даты/времени в строку.
    """
    description = 'Перевод даты/времени в строку'
    args_description = [
        MandatoryArg('Дата', 0, [datetime, date, time]),
        NonMandatoryArg('Формат', 1, [str]),
    ]

    def _operation(self, *args):
        dt = args[0]
        try:
            fmt = args[1]
        except IndexError:
            fmt = _DEFAULT_FMTS.get(type(dt), _DEFAULT_FMT)
        return dt.strftime(fmt)


class StrToDate(DateFunc):
    """
    Перевод строки в дату/время.
    """
    description = 'Перевод строки в дату/время'
    args_description = [
        MandatoryArg('Строка', 0, [str]),
        NonMandatoryArg('Формат', 1, [str]),
    ]

    def _operation(self, *args):
        dt = args[0]
        try:
            fmt = args[1]
        except IndexError:
            fmt = _DEFAULT_FMT
        return datetime.strptime(dt, fmt)



"""
Парсинг функций и формул.
"""

import re
from abc import ABC, abstractmethod

from .errors import ExpressionParsingError


def parse(formula_string):
    lexer = FormulaLexer(formula_string)
    tokens = [tok for tok in lexer.lex()]
    token_stream = Stream(tokens)
    parser = Parser(token_stream)
    stree = [tok for tok in parser.parse()]
    return stree


class Parser:
    """
    Парсер.
    """
    def __init__(self, token_stream, stop_at=';'):
        self._tokens = token_stream
        self._stop_at = stop_at

    @property
    def stop_at(self):
        return self._stop_at

    @property
    def tokens(self):
        return self._tokens

    def parse(self):
        while self.tokens.next is not None:
            e = self.next_expression(None)
            if e is not None:
                yield e
            self.tokens.get_next()

    def next_expression(self, prev):
        next_ = self.tokens.next
        type_, value = next_['type'], next_['value']
        if type_ in self.stop_at:
            return prev
        self.tokens.get_next()
        if type_ in ('number', 'string', 'symbol') and prev is None:
            return self.next_expression(next_)
        elif type_ == 'operator':
            nxt = self.next_expression(None)
            return self.next_expression({'type': 'operator', 'value': value, 'left': prev, 'right': nxt})
        elif type_ in '([':
            if type_ == '(':
                args = self._scan_args(func=prev)
                return self.next_expression({'type': 'call', 'function': prev, 'args': args})
            elif type_ == '[':
                args = self._scan_args(func=prev, end=']')
                return self.next_expression({'type': 'context', 'var': prev, 'args': args})
        else:
            raise ExpressionParsingError(f'Неожиданный токен: {(type_, value)}')

    def _scan_args(self, sep=',', end=')', func=None):
        args = []
        type_ = self.tokens.next['type']
        if type_ == end:
            self.tokens.get_next()
        else:
            if func is None:
                stop_at = (end)
            else:
                stop_at = (sep, end)
            arg_parser = Parser(self.tokens, stop_at)
            while type_ != end:
                p = arg_parser.next_expression(None)
                if p is not None:
                    args.append(p)
                type_ = self.tokens.next['type']
                self.tokens.get_next()
        return args


class Stream:
    def __init__(self, tokens):
        self._stream = iter(tokens)
        self._next = None
        self.set_next()

    def set_next(self):
        try:
            self._next = next(self._stream)
        except StopIteration:
            self._next = None

    def get_next(self):
        n = self.next
        self.set_next()
        return n

    @property
    def next(self):
        return self._next


class FormulaLexer:
    """
    Лексер для формул.
    """
    def __init__(self, tokens):
        self._rules = [
            NumberRule(),
            SymbolRule(),
            OperatorRule(),
            StringRule(),
            ErrorRule(),
        ]

        self._stream = Stream(tokens)

    def lex(self):
        stream = self._stream
        ready_for_next = True
        while stream.next is not None:
            if ready_for_next:
                token = stream.get_next()
            else:
                token = stream.next
            if token in ' \n':
                ready_for_next = True
            elif token in '(),;[]':
                ready_for_next = True
                yield {'type': token, 'value': ''}
            for rule in self._rules:
                if rule.match(token):
                    value, ready_for_next = rule.scan(stream, token)
                    yield {'type': rule.type, 'value': value}
                    break


class Rule(ABC):
    """
    Абстрактное правило.
    """
    @abstractmethod
    def match(self, token):
        pass

    @abstractmethod
    def scan(self, token_stream, token):
        pass

    @property
    @abstractmethod
    def type(self):
        return


class SymbolRule(Rule):
    """
    Правило для символов.
    """
    def match(self, token):
        if re.match('[a-zA-Z_0-9]', token):
            return True

    def scan(self, token_stream, token):
        res = token
        c = token_stream.next
        while c is not None and re.match(self.scan_re, c):
            res += token_stream.get_next()
            c = token_stream.next
        return res, True

    @property
    def scan_re(self):
        return '[a-zA-Z_0-9]'

    @property
    def type(self):
        return 'symbol'


class StringRule(Rule):
    """
    Правило для строк.
    """
    def match(self, token):
        if token in ['"', "'"]:
            return True

    def scan(self, token_stream, token):
        delimiter = token
        res = ""
        while token_stream.next != delimiter:
            c = token_stream.get_next()
            if c is None:
                raise ExpressionParsingError('Незавершённый идентификатор строки')
            res += c
        token_stream.get_next()
        return res, True

    @property
    def scan_re(self):
        return

    @property
    def type(self):
        return 'string'


class NumberRule(Rule):
    """
    Правило для чисел.
    """
    def match(self, token):
        if re.match('[0-9.]', token):
            return True

    def scan(self, token_stream, token):
        res = token
        c = token_stream.next
        while c is not None and re.match(self.scan_re, c):
            res += token_stream.get_next()
            c = token_stream.next
        return res, True

    @property
    def scan_re(self):
        return '[0-9.]'

    @property
    def type(self):
        return 'number'


class OperatorRule(Rule):
    """
    Правило для операторов.
    """
    def match(self, token):
        if token in ['+', '-', '*', '/', '>', '<', '!', '=', '&', '|']:
            return True

    def scan(self, token_stream, token):
        if token in ('>', '<', '!', '='):
            c = token_stream.get_next()
            if c == '=':
                return token + c, True
            else:
                if token in ('!', '='):
                    raise ExpressionParsingError(f'Неожиданный токен после "{token}": {c}')
                return token, True
        return token, True

    @property
    def scan_re(self):
        return

    @property
    def type(self):
        return 'operator'
            

class ErrorRule(Rule):
    def match(self, token):
        if not re.match('[ \n0-9a-zA-Z_()\[\];.,+-/*]', token):
            raise ExpressionParsingError(f'Неизвестный токен: {token}')

    def scan(self, token_stream, token):
        pass

    @property
    def type(self):
        pass